/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>
#include <string>
#include <list>
#include <errno.h>

#include "misc.h"
#include "tfile.h"

using namespace std;

namespace misc
{

// "lstat" MUST be used here because a link
// returns the statbuf of its target with "stat"

/******************************************
 * Return TRUE if pathname is a directory *
 ******************************************/
gboolean isdir(char *fname)
{
struct stat64 statbuf;
if (!(lstat64(fname, &statbuf)))
  {
  if (S_ISDIR(statbuf.st_mode)) return(TRUE);
  }
return(FALSE);
}

/*********************************************
 * Return TRUE if pathname is a regular file *
 *********************************************/

gboolean isfile(char *pathname)
{
  struct stat64 statbuf;
  // Check overflow error needed for large files
  if (!(lstat64(pathname, &statbuf)))
    {
    if (S_ISREG(statbuf.st_mode)) return(TRUE);
    }

  return(FALSE); // Not a regular file or ERROR
}

/*************************************
 * Return TRUE if pathname is a link *
 *************************************/

gboolean islink(char *pathname)
{
struct stat64 statbuf;

if (! (lstat64(pathname, &statbuf)) )
  {
  if (S_ISLNK(statbuf.st_mode)) return(TRUE);
  }
return(FALSE); // Not a link or ERROR
}


/*************************************************
 * Return TRUE if pathname is a character device *
 *************************************************/
gboolean ischardev(char *pathname)
{
struct stat64 statbuf;

if (! (lstat64(pathname, &statbuf)) )
  {
  if (S_ISCHR(statbuf.st_mode)) return(TRUE);
  }
return(FALSE); // Not a char dev or ERROR
}

/*********************************************
 * Return TRUE if pathname is a block device *
 *********************************************/
gboolean isblockdev(char *pathname)
{
struct stat64 statbuf;

if (! (lstat64(pathname, &statbuf)) )
  {
  if (S_ISBLK(statbuf.st_mode)) return(TRUE);
  }
return(FALSE); // Not a block dev or ERROR
}

/*************************************
 * Return TRUE if pathname is a fifo *
 *************************************/
gboolean isfifo(char *pathname)
{
struct stat64 statbuf;

if (! (lstat64(pathname, &statbuf)) )
  {
  if (S_ISFIFO(statbuf.st_mode)) return(TRUE);
  }
return(FALSE); // Not a fifo or ERROR
}

/***************************************
 * Return TRUE if pathname is a socket *
 ***************************************/
gboolean issocket(char *pathname)
{
struct stat64 statbuf;

if (! (lstat64(pathname, &statbuf)) )
  {
  if (S_ISSOCK(statbuf.st_mode)) return(TRUE);
  }
return(FALSE); // Not a socket or ERROR
}

/*********************************************
 * Return TRUE if pathname is a special file *
 *********************************************/
gboolean isnode(char *pathname)
{
if (ischardev(pathname)) return TRUE;
else if (isblockdev(pathname)) return TRUE;
else if (isfifo(pathname)) return TRUE;
else if (issocket(pathname)) return TRUE;

return FALSE;
}

/**********************************************
 * Return TRUE if pathname name exists at all *
 **********************************************/
gboolean isname(char *pathname)
{
if (isfile(pathname)) return TRUE;
else if (isdir(pathname)) return TRUE;

return (isnode(pathname));
}
 

/**************************************
 * Return the mode as an octal string *
 * or empty if error                  *
 **************************************/
char *get_mode(char *pathname)
{
  struct stat64 statbuf;

  static char mode[6];

  if (!(lstat64(pathname, &statbuf)))
    {
      int s = 0;
      // uid, gid, sticky
      //mode[0] = char( ((statbuf.st_mode & 0x7000)>>9) + '0');

      if (statbuf.st_mode & S_ISUID) s += 4;
      if (statbuf.st_mode & S_ISGID) s += 2;
      if (statbuf.st_mode & S_ISVTX) s += 1;
      mode[0] = '0' + s;

      // owner 
      mode[1] = char( ((statbuf.st_mode & S_IRWXU)>>6) + '0');

      // group
      mode[2] = char( ((statbuf.st_mode & S_IRWXG)>>3) + '0');

      // others
      mode[3] = char( (statbuf.st_mode & S_IRWXO) + '0');

      mode[4] = 0;

    }
  else *mode = 0;

  return mode;
}

/*****************************
 * Convert hex string to int *
 * Ignore leading space.     * 
 *****************************/
unsigned int hex2ui(char *str)
{
  char *p = str;
  char d;        // value of a hex digit
  unsigned int val = 0;   // running total

  while (isblank(*p)) p++;

  while ( isxdigit(*p) )
    {
      d = *p++;
      d = toupper(d);
      if ((d >= 'A') && (d<='F')) d = d - 'A' + 10;
      else d -= '0';
      val = (val * 16) + d;
    }
  return val;
}

/********************************
 * Return length of file fname  *
 * in bytes.                    *
 * Return FALSE if it fails.  *
 ********************************/

gboolean get_fsize(char *fname, off64_t *size)
{
struct stat64 statbuf;

 if (!(lstat64(fname, &statbuf)))
    {
      *size = statbuf.st_size;
      return(TRUE);
    }
return(FALSE);
}

// Place hostname of the computer into hostname
void find_hostname(string *hostname, string *m)
{
  char name[512];
  string msg;
  
  if (gethostname(name, 512))
    {
      // ERROR
      *m = "Failed to obtain hostname : ";
      *m += strerror(errno);
      *m += " : using \"\"";
      //m->process_warning(msg);

      *hostname = "";
    }
  else *hostname = name;
}


// Create a spinner (integer values)
GtkWidget *create_spinner(int s_start, int s_min, int s_max)
{
GtkAdjustment *spinner_adj;
spinner_adj = (GtkAdjustment *) gtk_adjustment_new (s_start, s_min, s_max, 1.0, 5.0, 5.0);
// creates the spinner, with no decimal places
return( gtk_spin_button_new (spinner_adj, 1.0, 0) );
}


// Create a style
GtkStyle *create_style(GdkColor fg, GdkColor text, GdkColor bg)
{
GtkStyle *defaultstyle;
GtkStyle *newstyle;
int i;

defaultstyle = gtk_widget_get_default_style();
newstyle = gtk_style_copy(defaultstyle);

// Copy for each state
for (i=0; i<5; i++)
  {
  newstyle->fg[i] = fg;
  newstyle->text[i] = text;
  newstyle->bg[i] = bg;
  }
return(newstyle);
}


void remove_leading_and_terminal_space(string &str)
{
// remove leading junk
while (isspace(str[0])) str = str.substr(1);

// remove trailing junk
while ( str.length() && isspace(str[str.length()-1]) )
  {
  str = str.substr(0, (str.length()-1) );
  }

}

// Create a string in the form hh:mm:ss from t
void time2string(time_t t, string &t_str)
{
string str;
int seconds;
int minutes;
int hours;

minutes = (int)t / 60;
seconds = (int)t % 60;

hours = minutes / 60;
minutes %= 60;

str = TFile::itoa(hours);
if (str.length() < 2) str = "0" + str;
t_str = str + ":";

str = TFile::itoa(minutes);
if (str.length() < 2) str = "0" + str;
t_str += str + ":";

str = TFile::itoa(seconds);
if (str.length() < 2) str = "0" + str;
t_str += str;

}

/*******************************************
 * int suficmp(char * string, char *suffix *
 * Compare the suffix "suffix" with the    *
 * terminal part of "string" and return as *
 * a normal case insensitive compare.      *
 *******************************************/
int suficmp(char *string, char *suffix)
{
int rc;
    
int len_suffix = strlen(suffix);
int len_string = strlen(string);
    
if ( (rc = len_string - len_suffix) >= 0 )
  {
  rc = strcasecmp(&string[rc], suffix);
  }
return(rc);
}

/*******************************************
 * int sufcmp(char * string, char *suffix  *
 * Compare the suffix "suffix" with the    *
 * terminal part of "string" and return as *
 * a normal case-sensitive compare.        *
 *******************************************/
int sufcmp(char *string, char *suffix)
{
int rc;

int len_suffix = strlen(suffix);
int len_string = strlen(string);

if ( (rc = len_string - len_suffix) >= 0 )
  {
  rc = strcmp(&string[rc], suffix);
  }
return(rc);
}

// For convenience
int STREQ(char *x, char *y)
{
return(strcmp(x,y)==0);
}

int STRIEQ(char *x, char *y)
{
return(strcasecmp(x,y)==0);
}

int STREQN(char *x, char *y, int n)
{
return(strncmp(x,y,n)==0);
}

int STRIEQN(char *x, char *y, int n)
{
return(strncasecmp(x,y,n)==0);
}

int SUFEQ(char *x, char *y)
{
return(sufcmp(x,y)==0);
}

int SUFIEQ(char *x, char *y)
{
return(suficmp(x,y)==0);
}

}   // end if namespace misc
