/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <iostream>
#include <utime.h>
#include <dirent.h>
#include <sys/types.h>
#include <pwd.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <stdlib.h>
#include <fcntl.h>
#include <fstab.h>
#include <mntent.h>

#include "backup.h"
#include "restore.h"
#include "conf.h"
#include "../common/misc.h"
#include "new_process.h"
#include "../common/tfile.h"
#include "../common/messages.h"
#include "../common/version.h"

using namespace std;
using namespace misc;

Backup::Backup(SlaveMessages *sm)
{
// cout << "Backup\n";
m = sm;
conf = NULL;
}

Backup::~Backup(void)
{
//  cout << "~Backup\n";

char uid[16];
char gid[16];
char mode[16];
char atime[16];
char mtime[16];

if (data.size())
  {
  list<File_Data>::iterator i;

  for (i=data.begin(); i!=data.end(); ++i)
    {

    sprintf(uid, "%X",  i->uid);
    sprintf(gid, "%X", i->gid);
    sprintf(mode, "%X", i->mode);
    sprintf(atime, "%08X", (unsigned)i->atime);  // atime is signed
    sprintf(mtime, "%08X", (unsigned)i->mtime);  // mtime is signed

    data_ofp << "ORPHAN " << "\"" << i->name << "\" "
       << uid <<  " "
       << gid <<  " "
       << mode <<  " "
       << atime <<  " "
       << mtime <<  " " << endl;

    }
  data.clear();
  }

if (conf)
  {
  delete conf;
  conf = NULL;
  }

if ( errors_ofp.is_open() )
  errors_ofp.close();
if (data_ofp.is_open() )
  data_ofp.close();
}



// Return TRUE if not forbidden by backup_deny
// backup_deny has precedence
// Also mustn't save destination to itself...
bool Backup::is_permitted(string *pathname)
{
list<string>::iterator i;
string msg;

// Destination is NOT ALLOWED!
if (STREQN((char *)pathname->c_str(),
        (char *)conf->backup_destination.c_str(),
        conf->backup_destination.length() ))
  {
  msg = "Backup of \"" + *pathname;
  msg += "\" not permitted (destination)";
  m->process_warning(msg);
  return FALSE;
  }

// There has to be a better way...
for (i=conf->backup_deny.begin(); i!=conf->backup_deny.end(); ++i)
  {
  if (STREQN((char *)pathname->c_str(), (char *)i->c_str(), i->length()))
    {
    msg = "Backup of \"" + *pathname;
    msg += "\" not permitted (omissions list)";
    m->process_warning(msg);
    return FALSE;   // e.g.  /proc/xxx is forbidden by /proc
    }
  }
return TRUE;
}


// Recursively delete all the files and directories
// It appears that permissions have to be changed...
bool Backup::delete_directory(string *dname)
{
DIR *dp;
struct dirent *de;
string name;
string msg;

//cout << "Backup::delete_directory - directory= " << *dname << endl;

// MAKE THE DIRECTORY READ/WRITE to ensure deletion
if (!(change_permissions_to_0777(dname, m)))
  {
  return FALSE;  // Failed so can't continue
  }

dp = opendir(dname->c_str());
if (dp)
  {
  while ( (de = readdir(dp)) )
    {
    if ( STREQ(de->d_name, ".") || STREQ(de->d_name, "..")) continue;
      
    name = *dname + "/";
    name += de->d_name;

    // cout << "Backup::delete_directory - got entry for " << name << endl;

    if (isdir((char *)name.c_str()))
      {      
      delete_directory(&name);
      }
    // Otherwise delete a different sort of name from the filesystem
    else
      {
      // MAKE THE NAME READ/WRITE to ensure deletion
      if (!(change_permissions_to_0777(&name, m)))
        {
        continue;  // Failed so try next name in the directory
        }

      if (remove(name.c_str()))
        {
        // ERROR
        msg = "Failed to delete filename \"";
        msg += name;
        msg += "\" : ";
        msg += strerror(errno);
        m->process_warning(msg);
        //cout << "Backup::delete_directory - " << msg << endl;
        }
      }
    }
    
  if (rmdir(dname->c_str()))
    {
    // ERROR
    msg = "Failed to delete directory \"";
    msg += *dname;
    msg += "\" : ";
    msg += strerror(errno);
    m->process_warning(msg);
    // cout << "Backup::delete_directory - " << msg << endl;
    return FALSE;
    }
  }
return TRUE;
}



void Backup::delete_old_backups(string *basename)
{
DIR *dp;
struct dirent *de;
string sname;
string pathname = conf->backup_destination;
string msg;

double stamp = 0;
char buffer[128];   // unnecessarily large!
char *b;            // position in buffer
char *p;            // position in current name being checked

list<BU_Data> bud;
list<BU_Data>::iterator i;
BU_Data entry;

int n = 0;          // number of backups in the list

if (conf->keep_number < 0) return;  // if 0, all backups are retained

if (pathname[pathname.length() - 1] != '/')
  {
  pathname += '/';
  }

dp = opendir(conf->backup_destination.c_str());
if (dp)
  {
  while ( (de = readdir(dp)) )
    {
    if ( STREQ(de->d_name, ".") || STREQ(de->d_name, "..")) continue;

    sname = pathname + de->d_name;

    if ( STREQN((char *)sname.c_str(),
          (char *)basename->c_str(),
          basename->length() ))
      {

      //cout << "Backup::delete_old_backups - GOT sname= " << sname << endl;

      // Got one, check it is for real
      p = &sname[basename->length()]; // skip over the basename
      b = buffer;


      if (strlen(p) != 19)
        {
        // BAD NAME
        //cout <<  "Backup::delete_old_backups - BAD NAME1 " << p << endl;
        continue;
        }

      while (*p)
        {
        if (*p == '_') {;} // skip a '_'
        else if ( (*p) && isdigit(*p) ) *b++ = *p;
        else
          {
          // BAD NAME
          //cout <<  "Backup::delete_old_backups - BAD NAME2 " << p << endl;
          break;
          }
        p++;
        }
      *b = 0;

      if ((strlen(buffer) != 14) || *p)
        {
        // BAD NAME
        //cout << "Backup::delete_old_backups - BAD NAME3 " << p << endl;
        continue;
        }

      stamp = atof(buffer);
      // Add this one to the list in order of age, oldest first

      for (i=bud.begin(); i!=bud.end(); ++i)
        {
        if ( stamp < (*i).stamp) break;
        }

      // Insert it here
      entry.name = sname;
      entry.stamp = stamp;

      bud.insert(i, entry);
        
      n++;
       
      if (n-1 > conf->keep_number)
        {    
        // if it is oldest, delete it
        // otherwise insert it and delete the oldest

        if (delete_directory( &(*( i=bud.begin())).name ) )
          {
          msg = "Deleted old backup: \"" + (*i).name;
          msg += "\"";
          m->log_message(msg);
          bud.erase(i);
          n--;
          }
        else
          {
          msg = "Failed to delete old backup: \"" + (*i).name;
          msg += "\" : ";
          msg += strerror(errno);
          m->process_warning(msg);
          // cout << "Backup::delete_old_backups - FAILED to remove" << (*i).name << endl;
          }
        }
      }
    }
  }

  for (i=bud.begin(); i!=bud.end(); ++i)
    {
    //      cout << ">>Backup::delete_old_backups - retained " <<  (*i).name << endl;

    msg = "Retained backup: \"" + (*i).name;
    msg += "\"";
    m->log_message(msg);
    }

bud.clear();
}

bool Backup::add_timestamp_to_pathname(string *pathname)
{
struct tm *lt;

char timestamp[1024];

lt = localtime(&m->start_time);
if (lt == NULL)
  {
  // ERROR!
  return FALSE;
  }

// Not sure whether to use this
// Could use t itself for a single number of seconds since epoch
sprintf(timestamp, "%04d_%02d_%02d_%02d_%02d_%02d",
    1900 + lt->tm_year,
    lt->tm_mon + 1,
    lt->tm_mday,
    lt->tm_hour,
    lt->tm_min,
    lt->tm_sec);

*pathname += timestamp;
return TRUE;
}


// Fix up a name to agree with sh if it contains a "'".
// The source is s which is put into name.
void Backup::fix_name(string &name, string *s)
{
unsigned int i = 0;
name = *s;

while (i < name.length())
  {
  if ( name[i] == '\'' )
    {
    // This needs to be changed to '\''
    // i.e. insert \'' after the first '
    //          or '\' before this character

    name.insert(i, "'\\'"); // escape_sequence);
    i += 3;
    }
  i++;
  }
} 

// Fix backup ownership and permissions of backup file
// Owner and group should be root, root can read and write, anyone else read only
// THIS FAILS!  WHY?
bool Backup::set_dest_attributes(string *dest, bool link)
{
string msg;

if (lchown(dest->c_str(), 0, 0))
  {
  // ERROR: Can't chown!  - possibly not important - IT IS IMPORTANT!
  msg = "Can't change ownership of \"" + *dest;
  msg += "\" to root : ";
  msg += strerror(errno);
  m->process_error(msg);

  //  cout << "Backup::set_dest_attributes - lchown FAILED\n";
  return FALSE;
  }

if (isdir((char *)dest->c_str()))
  {
  if (!change_permissions_to_0777(dest, m))
  
  if (chmod(dest->c_str(), 0777))
    {
    // cout << "Backup::set_dest_attributes - chmod FAILED\n";
    return FALSE;
    }
  }
else if ((!link) && (chmod(dest->c_str(), 0644)))
  {
  msg = "Can't change permissions of \"" + *dest;
  msg += "\" to 644 : ";
  msg += strerror(errno);
  m->process_error(msg);

  // cout << "Backup::set_dest_attributes - chmod FAILED\n";
  return FALSE;
  }
return TRUE;
}

// Check if in data list, remove if so and return TRUE else return FALSE
bool Backup::check_update_list(string *s_name)
{
if (update)
  {
  // If s_name is in the data list it has to be removed
  // This leaves orphans at the end
  list<File_Data>::iterator i;

  for (i=data.begin(); i!=data.end(); ++i)
    {
    if ( (*i).name == *s_name)
      {
      data.erase(i);
      return TRUE;
      }
    }
  }
return FALSE;
}

// Save source info
void Backup::save_attributes(string *s_name, struct stat64 *statbuf)
{
char uid[16];
char gid[16];
char mode[16];
char atime[16];
char mtime[16];

sprintf(uid, "%X", statbuf->st_uid);
sprintf(gid, "%X", statbuf->st_gid);
sprintf(mode, "%X", statbuf->st_mode);
sprintf(atime, "%08X", (unsigned)statbuf->st_atime); // atime is signed
sprintf(mtime, "%08X", (unsigned)statbuf->st_mtime); // mtime is signed

data_ofp << "\"" << *s_name << "\" "
        << uid << " "
        << gid << " "
        << mode << " "
        << atime << " "
        << mtime << endl;

check_update_list(s_name);

m->num_names++;      
}


bool Backup::create_directory(string *name)
{
  string msg;

  if (mkdir(name->c_str(), 0777))
    {
      // ERROR: Can't create!
      msg = "Failed to create directory \"";
      msg += *name;
      msg += "\" : ";
      msg += strerror(errno);

      m->process_error(msg);
      return FALSE;
    }

  return (set_dest_attributes(name, FALSE));

}






// The source is the file to be backed up
// The list file should be an earlier backup's original name and details
// Returns TRUE if name is in list
// If newer is TRUE the name has to be in the list!
bool Backup::name_in_list(string *src, bool &newer)
{
list<File_Data>::iterator i;

//  cout << "Backup::name_in_list - checking \"" << *src << "\"" << endl;

for (i=data.begin(); i!=data.end(); ++i)
  {
  if ( (*i).name == *src )
    {
    struct stat64 statbuf;
    if (!(lstat64(src->c_str(), &statbuf)))
      {
      // Greater is older than (seconds since epoch)
      if (statbuf.st_mtime > (*i).mtime)
        {
        newer = TRUE;
        }
      else
        {
        newer = FALSE;
        }
      }
    else
      {
      // ERROR
      // Assume the src file is newer if it can't be "lstat64"-ed
      // Perhaps it should write an error...
      newer = TRUE;
      }

    return TRUE;   // name was in list
    }
  }
newer = FALSE;
return FALSE;
}


bool Backup::change_permissions_to_0777(string *name, SlaveMessages *m)
{
  string msg;

  // NOTE: the 777 MUST be preceded by 0
  // This should make it fully accessible but NOT SO!
  if (chmod(name->c_str(), 0777)) // 0777))
    {
      // ERROR
      msg = "Can't change permissions of \"";
      msg += *name;
      msg += "\" to 0777 : ";
      msg += strerror(errno);
      m->process_error(msg);
      return FALSE;
    }
  return TRUE;
}

bool Backup::change_permissions_to_0755(string *name)
{
  string msg;

  if (chmod(name->c_str(), 0755))
    {
      // ERROR
      msg = "Can't change permissions of \"";
      msg += *name;
      msg += "\" to 0755 : ";
      msg += strerror(errno);
      m->process_error(msg);
      return FALSE;
    }
  return TRUE;
}



bool Backup::restore_permissions(string *name,
         struct stat64 *statbuf,
         SlaveMessages *m)
{
  string msg;

  if (chmod(name->c_str(), statbuf->st_mode))
    {
      msg = "Can't restore permissions for \"" + *name;
      msg += "\" : ";
      msg += strerror(errno);
      m->process_error(msg);
      return FALSE;
    }
  return TRUE;
}

bool Backup::get_statbuf(string *name, struct stat64 *statbuf, SlaveMessages *m)
{
  string msg;

  if (!(lstat64((char *)name->c_str(), statbuf)))
    {
      return TRUE;
    }

  // ERROR
  msg = "Can't lstat64 \"" + *name;
  msg += "\" : ";
  msg += strerror(errno);
  m->process_error(msg);
  return FALSE;
}



bool Backup::process_file(string *d_file, string *s_file, struct stat64 *statbuf)
{
  string command;
  string dest;
  string temp;
  string msg;
  bool inlist;
  bool newer;
  bool dest_exists;

  string d;
  string s;

  //cout << "Backup::process_file - FILE= \"" << *s_file << "\"\n";
  //cout << "Backup::process_file - MODE= " << (char *)get_mode((char *)s_file->c_str()) << endl;

  //cout << "          Source= " << *s_file << endl;
  //cout << "     Destination= " << *d_file << endl;

  if (compression == "GZ")
    {
      dest = *d_file + ".gz";
    }
  else if (compression == "BZ2")
    {  
      dest = *d_file + ".bz2";
    }
  else if (compression == "GZ_BZ2")
    {  
      dest = *d_file + ".gz.bz2";
    }
  else dest = *d_file;

  // Prepare name for use by shell
  fix_name(s, s_file);

  //cout << "Backup::process_file:-\n";
  //cout << "  s_file= " << *s_file << endl;
  //cout << "       s= " << s       << endl;

  // Prepare name for use by shell
  fix_name(d, &dest);



  // NOTE: if update, it can't be in the list because the list is empty

  inlist = name_in_list(s_file, newer);

  dest_exists = (update && isfile((char *)dest.c_str())) ? TRUE : FALSE;

  //cout << "Backup::process_file - DEST       = \"" << dest << "\"\n";
  //cout << "Backup::process_file - newer      = " << newer << " inlist= " << inlist << "\n";
  //cout << "Backup::process_file - dest_exists= " << dest_exists << endl;

  if (update && dest_exists && !newer)
    {
      // We don't want this one...
      // It needs to be removed from the data list, then return TRUE

      // msg = "SKIPPING \"";
      // msg += *s_file;
      // msg +=  "\", not newer";  
      // m->log_message(msg);

      // cout << "Backup::process_file - " << msg << endl;
      m->num_skipped++;
      return TRUE;
    }

  if ( (!inlist) || newer || (!(dest_exists)) )
    {
    if (update && newer && dest_exists)
      {
      // RENAME THE DESTINATION FILE

      TFile tf;
      string suffix = ".spiffasave";
      string name = dest + "_";

      temp = *tf.tname(&name, 3, &suffix);
      if (rename(dest.c_str(), temp.c_str()))
        {
        // ERROR - couldn't rename
        msg = "Failed rename pre-existing destination file \"";
        msg += dest;
        msg += "\" as \"";
        msg += temp;
        msg += "\" : ";
        msg += strerror(errno);
        m->process_error(msg);
        return FALSE;
        }
      }

    if (compression == "GZ")
      {
      command = conf->cat_pname + " '";
      command += s;
      command += "' | ";
      command += conf->gzip_pname;
      command += " -q -9 -N > '";
      command += d;
      command += "'";
      }
    else if (compression == "BZ2")
      {  
      command = conf->cat_pname + " '";
      command += s;
      command += "' | ";
      command += conf->bzip2_pname;
      command += " -k -q > '";
      command += d;
      command += "'";
      }
    else if (compression == "GZ_BZ2")
      {  
      command = conf->cat_pname + " '";
      command += s;
      command += "' | ";
      command += conf->gzip_pname;
      command += " -q -9 -N | ";
      command += conf->bzip2_pname;

      command += " -k -q > '";
      command += d;
      command += "'";
      }

    else  // DEFAULT - assume that (compression == "NONE")  
      {
      command = conf->cat_pname + " '";
      command += s;
      command += "' > '";
      command += d;
      command += "'";
    }

  /***
  { // cat is more useful than this
    command = "cp -P -p \"";
    command += *s_file;
    command += "\"  \"";
    command += *d_file;
    command += "\"";
  }
  ***/

  // cout << "Backup::process_file - command=\n     " << command << endl;

  if ( run.new_process(command, conf->sh_pname) )
    {
    // cerr << "Backup::process_file - command FAILED" << endl;
    
    // temp file should be re-named back to original
    //==============================================
    msg = "Command \"";
    msg += command;
    msg += "\" FAILED : ";
    msg += run.status_string;

    m->process_error(msg);

    if (newer && dest_exists)
      {
      if (rename(temp.c_str(), dest.c_str()))
        {
        // ERROR - couldn't rename
        msg = "Failed to rename temporary file \"";
        msg += temp;
        msg += "\" back to \"";
        msg += dest;
        msg += "\" : ";
        msg += strerror(errno); 
        m->process_warning(msg);
        }
      }   
    return FALSE;
    }

  // Check that the destination was created! 
  if (!isfile((char *)dest.c_str()))
    {
    // ERROR

    if (newer && dest_exists)
      {
      if (rename(temp.c_str(), dest.c_str()))
        {
        // ERROR - couldn't rename
        msg = "Failed to rename temporary file \"";
        msg += temp;
        msg += "\" back to \"";
        msg += dest;
        msg += "\" : ";
        msg += strerror(errno); 
        m->process_warning(msg);
        }
      }
    return FALSE;
    }
  else if (newer && dest_exists)
    {
    if (remove(temp.c_str()))
      {
      // ERROR
      msg = "Failed to delete temporary file \"";
      msg += temp;
      msg += "\" :";
      msg += strerror(errno);
      m->process_warning(msg);
      }
    }
  }

set_dest_attributes(&dest, FALSE);
return TRUE;
}


// Get the target of a link
bool Backup::get_target(string *name, string &target, SlaveMessages *m)
{
struct stat64 statbuf;
long int len;
string msg;

if (!Backup::get_statbuf(name, &statbuf, m))
  {
  return FALSE;
  }

if (!S_ISLNK(statbuf.st_mode))
  {
  msg = "\"";
  msg += *name;
  msg += "\" is not a link!";
  m->process_error(msg);
  return FALSE;
  }
else
  {
  char name_target[statbuf.st_size + 1];

  // This returns the correct length of the link target even if broken
  len = statbuf.st_size;

  // This always returns 0 unless error
  if ( (readlink(name->c_str(), name_target, len) == -1) )
    {
    msg = "Failed to read link \"";
    msg += *name;
    msg += "\" : ";
    msg += strerror(errno);
    m->process_error(msg);
    return FALSE;
    }
  name_target[len] = 0;
  target = name_target;
  return TRUE;
  }
}





// LINK
// A link can be renamed
// BUT it appears that chown and chmod work on the target, not the link itself
// USE lchown to change the ownership of a link - works for all files...
// Can't use chmod for links - there is no lchmod.
// If the link already exists in the destination, could check that
// it points to the same pathname and then accept that in an update.
// Probably better for update to update all links
bool Backup::process_link(string *d_link, string *s_link, struct stat64 *statbuf)
{
string command;
string temp;
string msg;

bool inlist;
bool newer;
bool dest_exists;
string s_target;

string t;
string d;

// cout << "Backup::process_link - s_link= " << *s_link
//     << "\n                     - dlink= " << *d_link << endl;

if (!(get_target(s_link, s_target, m))) return FALSE;

fix_name(t, &s_target);
fix_name(d, d_link);

// cout << "Backup::process_link - target= " << target << endl;


// NOTE: if update, it can't be in the list because the list is empty
inlist = name_in_list(s_link, newer);


dest_exists = (update && islink((char *)d_link->c_str())) ? TRUE : FALSE;

/**************************************************************
if (dest_exists)
  {
  string d_target;
  if (!(get_target(d_link, d_target, m))) return FALSE;

  if (d_target == s_target)
    {
    //cout << "Backup::process_link - TARGETS THE SAME" << endl;
    // We don't want this one...
    // It needs to be removed from the data list, then return TRUE
    msg = "SKIPPING \"";
    msg += *s_link;
    msg +=  "\", targets are the same";  
    m->log_message(msg);
    return TRUE;
    }
  //else  cout << "Backup::process_link - TARGETS DIFFERENT" << endl;
  }
****************************************************************/



/**************************************************************************
cout <<             "Backup::process_link - s_link= " << *s_link << endl;
if (inlist) cout << "      IN LIST         - dlink= " << *d_link;
else        cout << "    NOT IN LIST       - dlink= " << *d_link;
cout << (char *)((newer) ? " NEWER" : " OLDER");
if (dest_exists) cout << " : DEST EXISTS" << endl;
else             cout << " : DEST MISSING" << endl;
**************************************************************************/

if (newer || (!inlist) || (!(dest_exists)) )
  {
  if (update && newer && dest_exists)
    {
    // RENAME THE DESTINATION FILE

    TFile tf;
    string suffix = ".spiffasave";
    string name = *d_link + "_";

    temp = *tf.tname(&name, 3, &suffix);
    if (rename(d_link->c_str(), temp.c_str()))
      {
        // ERROR - couldn't rename
        msg = "Failed to rename pre-existing destination link \"";
        msg +=  *d_link;
        msg += "\" as \"";
        msg += temp;
        msg += "\" : ";
        msg += strerror(errno);
        m->process_error(msg);
        return FALSE;
      }
    }

  command = conf->ln_pname + " -s \'";
  command += t;           // the target
  command += "\' \'";
  command += d;           // the link name that is created
  command += "\'";

  if (run.new_process(command, conf->sh_pname))
    {
    msg = "Command \"";
    msg += command;
    msg += "\" FAILED : ";
    msg += run.status_string;
    m->process_error(msg);

    if (newer && dest_exists)
      {
      if (rename(temp.c_str(), d_link->c_str()))
        {
        // ERROR
        msg = "Failed to rename temporary link \"";
        msg += temp;
        msg += "\" back to \"";
        msg += *d_link;
        msg += "\" : ";
        msg += strerror(errno);
        m->process_error(msg);
        }
      }
    return FALSE;
    }

  if (!islink((char *)d_link->c_str())) // If run succeded it should be present!
    {
    // ERROR
    msg = "Failed to create destination link \"";
    msg += *d_link;
    msg += "\" although command to do so succeded! : ";
    msg += strerror(errno);
    m->process_error(msg);

    if (newer && dest_exists)
      {
      if (rename(temp.c_str(), d_link->c_str()))
        {
        // ERROR
        msg = "Failed to rename temporary link \"";
        msg += temp;
        msg += "\" back to \"";
        msg += *d_link;
        msg += "\" : ";
        msg += strerror(errno);
        m->process_error(msg);
        }
      }
    return FALSE;
    }
  else if (newer && dest_exists)
    {
    if (remove(temp.c_str()))
      {
      // ERROR
      msg = "Failed to delete temporary link \"";
      msg += temp;
      msg += "\" : ";
      msg += strerror(errno);
      m->process_warning(msg);
      }
    }
  }

// Unfortunately chmod sets the link target, not the link itself
// This uses lchown for ownership, where is lchmod?
set_dest_attributes(d_link, TRUE); 

return TRUE;
}


// ======================================================================
// The code for devices, fifos and sockets is almost the same
// Only the mknode function is different, so pass an int 0, 1 or 2
// for device, fifo or socket respectively
// then do a switch
// These special files are always updated by update
// ======================================================================
bool Backup::process_node(string *d_node, string *s_node, struct stat64 *statbuf)
{
string temp;
string msg;
bool inlist;
bool newer;
bool dest_exists;
int rc;

inlist = name_in_list(s_node, newer);

dest_exists = (update && isnode((char *)d_node->c_str()));


if ( newer || (!inlist) || (!(dest_exists)) )
  {
  if (update && newer && dest_exists)
    {
    // RENAME THE DESTINATION FILE

    TFile tf;
    string suffix = ".spiffasave";
    string name = *d_node + "_";

    temp = *tf.tname(&name, 3, &suffix);
    if (rename(d_node->c_str(), temp.c_str()))
      {
      // ERROR - couldn't rename
      msg = "Failed to rename pre-existing destination special file \"";
      msg +=  *d_node;
      msg += "\" as \"";
      msg += temp;
      msg += "\" : ";
      msg += strerror(errno);
      m->process_error(msg);
      return FALSE;
      }
    }

  if ( (S_ISCHR(statbuf->st_mode)) || (S_ISBLK(statbuf->st_mode)) )
    {
    rc = mknod(d_node->c_str(), statbuf->st_mode, statbuf->st_rdev);
    }
  else if (S_ISFIFO(statbuf->st_mode))
    {
    rc = mkfifo(d_node->c_str(), statbuf->st_mode);
    }
  else if (S_ISSOCK(statbuf->st_mode))
    {
    rc = mknod(d_node->c_str(), statbuf->st_mode, 0);
    }
  else rc = -2;

  if (rc)
    {
    // ERROR
    msg = "Failed to ";
    msg += ((update) ? "update" : "backup");
    msg += " special file \"";
    msg += *s_node;
    msg += "\" : ";
    if (rc == -2 )msg += "UNKNOWN FILE TYPE";
    else msg += strerror(errno);
    m->process_error(msg);

    if (newer && dest_exists)
      {
      if (rename(temp.c_str(), d_node->c_str()))
        {
        // ERROR
        msg = "Failed to rename temporary special file \"";
        msg += temp;
        msg += "\" back to \"";
        msg += *d_node;
        msg += "\" : ";
        msg += strerror(errno);
        m->process_error(msg);
        }
      }
    return FALSE;
    }

  if (newer && dest_exists)
    {
    if (remove(temp.c_str()))
      {
      // ERROR
      msg = "Failed to delete temporary special file \"";
      msg += temp;
      msg += "\" : ";
      msg += strerror(errno);
      m->process_warning(msg);
      }
    }

  set_dest_attributes(d_node, FALSE);
  }
return TRUE;
}



bool Backup::process_dir_contents(string *dir)
{
DIR *dp;
struct dirent *de;
string s_name;
string d_name;
bool rc;

// cout << "Backup::process_dir_contents - directory " << *dir << endl;

dp = opendir(dir->c_str());
if (dp)
  {
  // cout << "Backup::process_dir_contents - opened directory " << *dir << endl;
  
  while ( (de = readdir(dp)) )
    {
    if ( STREQ(de->d_name, ".") || STREQ(de->d_name, "..")) continue;
      
    s_name = *dir;
    if (*dir != "/") s_name += "/";
    s_name += de->d_name;
    d_name = root_pname;
    if (s_name[0] != '/') d_name += "/";
    d_name += s_name;

    // cout << "Backup::process_dir_contents - GOT           " << de->d_name << endl;
    // cout << "                             - GOT ENTRY FOR "
    //     << s_name << endl;

    rc = process_name(d_name, s_name, de->d_name);

    // cout << "Backup::process_dir_contents -     PROCESSED "
    //     << s_name <<  " " << ((rc) ? "SUCCESS" : "FAILURE") << endl;

    }
  closedir(dp);
  return TRUE;
  }
return FALSE;
}

// =================================================
// If updating, the destination has to be read/write
// =================================================
bool Backup::process_name(string &dest, string &source, char *filename)
{
// cout << "Backup::process_name - dest= " << dest << endl;

string msg;
struct stat64 statbuf;
bool rc = TRUE;     // assume success

m->get_messages();
if (m->quit_flag) return FALSE;


if (!(is_permitted(&source))) return FALSE;

// If no statbuf it can't go any further
if (!get_statbuf(&source, &statbuf, m)) return FALSE;


// MUST be readable
// For a link this changes permissions of its target, NOT THE LINK!
// Unfortunately a link target doesn't get restored properly
// WHY NOT?
if (!S_ISLNK(statbuf.st_mode))
  {
  if (!(change_permissions_to_0777(&source, m))) return FALSE;
  }

if (S_ISDIR(statbuf.st_mode))
  {
  m->send_message(SL_PATH, (char *)source.c_str());

  // cout << "Backup::process_name - source " << source << " is DIRECTORY" << endl;
      
  if (!isdir((char *)dest.c_str()))
    {
    if (!create_directory(&dest))
      {
        // ERROR: Can't create!
        rc = FALSE;
      }
    }

  // TRY THIS
  if ( rc && (update) && (!change_permissions_to_0777(&dest, m)) )
    {
    rc = FALSE;
    }

  if ( (rc==TRUE) && (!process_dir_contents(&source)) )
    {
    // ERROR
    msg = "FAILED to backup contents of directory \"";
    msg += source;
    msg += "\"";
    m->process_error(msg);
    rc = FALSE;
    }

  // Make the destination directory read only here
  rc = change_permissions_to_0755(&dest);
  }
else
  {
  m->send_message(SL_FILE, filename);

  if  (S_ISREG(statbuf.st_mode))
    {

    // cout << "Backup::process_name - source " << source << " is FILE" << endl;

    if (!process_file(&dest, &source, &statbuf))
      {
      // ERROR

      msg = "FAILED backup of file \"";
      msg += source;
      msg += "\"";
      m->process_error(msg);
      rc = FALSE;
      }
    }
  else if (S_ISLNK(statbuf.st_mode))
    {
    //  cout << "Backup::process_name - source " << source << " is LINK" << endl;

    if (!process_link(&dest, &source, &statbuf))
      {
      // ERROR
      msg = "FAILED backup of link \"";
      msg += source;
      msg += "\"";
      m->process_error(msg);
      rc = FALSE;
      }
    }

  // Anything else is either a node or an error
  else if (!process_node(&dest, &source, &statbuf))
    {
    // ERROR
    msg = "FAILED backup of special file \"";
    msg += source;
    msg += "\"";
    m->process_error(msg);
    rc = FALSE;
    }

  }


if (!S_ISLNK(statbuf.st_mode))
  {
  restore_permissions(&source, &statbuf, m); // ALWAYS restore permissions (except links)
  }

if (rc) // save if successful only
  {
  save_attributes(&source, &statbuf);
  }
return rc;
}

// Recursive version
// Process intermediate directories then the terminal file or directory
bool Backup::process_source(string &source, string *current)
{
char path[source.length() + 1];
char *p;

string command;
string dest_dir;
string src_dir;

if (m->quit_flag) return FALSE;

strcpy(path, source.c_str());

if (current) p = &path[current->length()];
else p = path;

p++;

while ( (*p) && ((*p)!='/') ) p++;

if ((*p) == 0)
  {
  // Terminal name

  string dest = root_pname + source;
  string filename;

  //cout << "Backup::process_source - " << source << " mode " << get_mode((char *)source.c_str()) << endl;

  process_name(dest, source, (char *)(get_filename(source, filename)).c_str() );
  }
else
  {
  // Intermediate directory
  struct stat64 statbuf;

  *p = 0;  // Only process this intermediate directory

  if ( !(isdir(path)) )   // If this is not a directory (it should be!)
    {
    // ERROR
    string msg = "Source \"" + source;
    msg += "\" includes non-existent directory \"";
    msg += path;
    msg += "\" : NOT PROCESSED.";
    m->process_error(msg);
    return FALSE;
    }

  src_dir = path;
  dest_dir = root_pname + path;

  if (!get_statbuf(&src_dir, &statbuf, m))
    {
    // ERROR
    return FALSE;
    }
    

  // Get frontend to display this path
  m->send_message(SL_PATH, path);

  // If update, check for this path in the data
  // If present, remove it and increment num_names
  if (check_update_list(&src_dir))
    {
    // Try save_attributes instead...
    save_attributes(&src_dir, &statbuf);
    }

  // Is a new destination directory required?
  if (!isdir((char *)dest_dir.c_str())) // if it doesn't already exist
    {   
    if (!create_directory(&dest_dir))   // Create a new destination directory
      {
      // ERROR  
        return FALSE;   // Failed so can't continue
        }
    save_attributes(&src_dir, &statbuf); // Write new directory data
    }

  // MAKE THE DIRECTORY READABLE
  if (!(change_permissions_to_0777(&src_dir, m)))
    {
    return FALSE;  // Failed so can't continue
    }

  if (!process_source(source, &src_dir))
    {
    // ERROR
    // ALWAYS FIX THE DIRECTORY PERMISSIONS BEFORE LEAVING!
    restore_permissions(&src_dir, &statbuf, m);
    return FALSE;   // Failed so can't continue
    }

  // ALWAYS FIX THE DIRECTORY PERMISSIONS BEFORE LEAVING!
  restore_permissions(&src_dir, &statbuf, m);
  }

return TRUE;
}




void Backup::process_source_list(void)
{
list<string>::iterator i;
string source;
string msg;

if (conf->backup_sources.empty())
    {
    msg = "Can't backup because source list is empty!";
    m->process_fatal_error(msg);
    return;
    }

for (i=conf->backup_sources.begin(); i!=conf->backup_sources.end(); ++i)
    {
    if (m->quit_flag) break;

    source = *i;

    // skip root. This is assumed...
    // sources should not be relative,
    // although the relative path could be obtained if necessary
    if (source[0] != '/')
        {

        // ============================
        // ERROR - path is not absolute
        // ============================
        msg = "Source name \"" + source;
        msg += "\" is not an absolute pathname: NOT PROCESSED.";
        m->process_error(msg);
        continue;
        }


    // The source must NOT end with a slash unless it is "/"
    while ((source.length()>1) && (source[source.length()-1]=='/'))
        {
        source = source.substr(source.length()-1);
        }

    if (is_permitted(&source))
        {

        if (!process_source(source, NULL))
            {
            // ERROR
      
            // cerr << "Backup::new_backup - source " << source << " - process_source FAILED" << endl;
            }
        }
    }

// Fix up root then backup here to 0755
change_permissions_to_0755(&root_pname);   // root
change_permissions_to_0755(&pathname);     // backup

if ( (!m->quit_flag) && (m->get_num_errors() == 0) )
    {
    // Delete the old backupswant_
    delete_old_backups(&basename);
    }
}


// ================================================
// Check that mount point is OK if it was specified
// ================================================
bool Backup::mount_point_OK(void)
{
string msg;
struct fstab *fs;

if (conf->mount_point.length() == 0) return TRUE; // no mp specified

// remove trailing slashes
while ( (conf->mount_point.length()>1) && (conf->mount_point[conf->mount_point.length()-1]=='/') )
  {
  conf->mount_point = conf->mount_point.substr(0, (conf->mount_point.length()-1) );
  }

if (conf->backup_destination != conf->mount_point)
  {
  string mp = conf->mount_point + '/';

  if (!STREQN((char *)conf->backup_destination.c_str(), (char *)mp.c_str(), mp.length()))
    {
    // ERROR
    msg += " Can't perform ";
    msg += (update) ? "update" : "new backup. ";
    msg = "Mount Point \"" + conf->mount_point;
    msg += "\" conflicts with destination \"";
    msg += conf->backup_destination;
    msg += "\"";
    m->process_fatal_error(msg);
    return FALSE;
    }
  }


// Is this mount point in fstab?
setfsent();
while ( (fs=getfsent()) )
  {
//  cout << "Checking \"" << fs->fs_file << "\"" << endl;
  // FOUND IT
  if (fs->fs_file == conf->mount_point)
    {
    break;
    }
  }
endfsent();

/****************************
if (fs == NULL)
  {
  // FAILED to find this mount point in fstab,
  // might or might not be OK.
  }
****************************/

// If the mount point directory exists
// check if it is in mtab.
// If so, it is mounted.
if (isdir((char *)conf->mount_point.c_str()))
  {
  FILE *fp;
  struct mntent *mnt;

  fp = setmntent((char *)conf->mtab_pname.c_str(), "r");
  if (fp == NULL)
    {
    // ERROR
    msg += " Can't perform ";
    msg += (update) ? "update" : "new backup. ";
    msg += "Unable to open \"";
    msg += conf->mtab_pname;
    msg += "\" for read so unable to check if directory \"";
    msg += conf->mount_point;
    msg += "\" is mounted";
    m->process_fatal_error(msg);
    return FALSE;
    }
  else
    {
    while ( (mnt = getmntent(fp)) )
      {
      if (conf->mount_point == mnt->mnt_dir)
        {
        endmntent(fp);
        return TRUE;
        }
      }
    }
  endmntent(fp);
  return FALSE;
  }

if (fs) // The mount point was found in fstab but no corresponding directory
  {
  msg = "Mount Point \"" + conf->mount_point;
  msg += "\" was found in \"fstab\" but there is no corresponding directory";
  m->process_warning(msg);
  }

return TRUE;
}

bool Backup::fix_conf(string *conf_name)
{
string hostname;
string msg;

// cout <<  "Backup::fix_conf\n";

umask(0777);  // make sure we can create all the modes we want


if (conf_name && conf_name->length())
  {
  conf = new Conf(*conf_name);    
  }
else conf = new Conf();
  
 
// Set up the destination directory
// The name should be /path/spiffa_backup_year_month_day_hour_minute_second
pathname = conf->backup_destination;

if (pathname.length() == 0)
  {
  // ERROR - no destination
  msg = "No destination specified!";
  msg += " Can't perform ";
  msg += (update) ? "update" : "new backup";
  m->process_fatal_error(msg);
  return FALSE;
  }

if (!isdir((char *)pathname.c_str()))
  {
  msg = "Destination \"" + pathname;
  msg += "\" doesn't exist! Can't perform ";
  msg += (update) ? "update" : "new backup";
  m->process_fatal_error(msg);
  return FALSE;
  }

if (!mount_point_OK())
  {
  return FALSE;
  }

misc::find_hostname(&hostname, &msg);


// If pathname doesn't end in a slash (/) add a slash
if (pathname[pathname.length() - 1] != '/')
  {
  pathname += '/';
  }

pathname += "spiffa_backup_";
pathname += hostname;
pathname += "_";
basename = pathname;

if (update)
  {
  if (!Restore::find_backup(&pathname, conf, m))
    {
    msg = "Can't find a backup to update";

    // cout << "Backup::fix_conf - " << msg << endl;;
    m->process_warning(msg);
    return FALSE;
    }
  root_pname = pathname + "/root";
  errors_pname = pathname + "/errors";
  data_pname = pathname + "/data";
  }
else
  {
  if (!add_timestamp_to_pathname(&pathname))
    {
    // ERROR! - no timestamp...
    msg = "Can't add the timestamp to the pathname";
    m->process_fatal_error(msg);

    return FALSE;
    }

  root_pname = pathname + "/root";
  errors_pname = pathname + "/errors";
  data_pname = pathname + "/data";

  if (!create_directory(&pathname))
    {
    msg = "Can't create directory \"" + pathname;
    msg += "\" : Can't perform new backup";
    m->process_fatal_error(msg);
    return FALSE;
    }

  if (!create_directory(&root_pname))
    {
    msg = "Can't create directory \"" + root_pname;
    msg += "\" : Can't perform new backup";
    m->process_fatal_error(msg);
    return FALSE;
    }
  }

// Make sure the restore point is fully accessible
if (!(change_permissions_to_0777(&pathname, m)))
  {
  msg = "Can't make directory \"" + pathname;
  msg += "\" fully accessible";
  m->process_fatal_error(msg);
  return FALSE;
  }
if (!(change_permissions_to_0777(&root_pname, m)))
  {
  msg = "Can't make directory \"" + root_pname;
  msg += "\" fully accessible";
  m->process_fatal_error(msg);
  return FALSE;
  }

// cout << "Backup::fix_conf - pathname is \"" << pathname << "\"\n";

msg = "Destination: " + pathname;
m->send_message(SL_POINT, (char *)msg.c_str());

msg = "Destination: " + pathname;
msg += "\n";
m->log_message(msg);

// Create file "errors"
if (update)
  {
  errors_ofp.open(errors_pname.c_str(), ios::out | ios::binary | ios::app );
  }
else
  {
  errors_ofp.open(errors_pname.c_str(), ios::out | ios::binary | ios::trunc );
  }
if ( ! errors_ofp.is_open() )
  {
  msg = "Can't open \"";
  msg += errors_pname;
  msg += "\" for write : ";
  msg += strerror(errno);
  m->log_message(msg);
  return FALSE;
  }
else
  {
  // Errors and data should be readable by all...
  chmod(errors_pname.c_str(), 0644);

  errors_ofp << "\n";
  errors_ofp <<  version_txt
              << ( (update) ? "\nUPDATE ERRORS " : "\nBACKUP ERRORS " )
              << ctime(&m->start_time);
  errors_ofp << "User: ";
  errors_ofp << (getpwuid( getuid() ))->pw_name;
  errors_ofp << "\nDestination: " << pathname << endl;

  m->put_error_fp(&errors_ofp);
  }

// Read in file data for update
// Also read in orphans as if part of the list
if (update) get_file_data(pathname, data, TRUE, m);

// Create file "data" - if it exists it is over-written
data_ofp.open(data_pname.c_str(), ios::out | ios::binary | ios::trunc ); 
if ( ! data_ofp.is_open() )
  {
  // ERROR - can't create files_data file, this vital!
  msg = "Can't open \"";
  msg += data_pname;
  msg += "\" for write : ";
  msg += strerror(errno);
  m->process_fatal_error(msg);
  return FALSE;
  }
else
  {
  chmod(data_pname.c_str(), 0644);
  data_ofp << "# " << version_txt << "\n# FILE DATA " << ctime(&m->start_time) << endl;
  }
return TRUE;
}

string &Backup::get_filename(string &pathname, string &filename)
{
unsigned int p = pathname.length();
filename.clear();

if (!isdir((char *)pathname.c_str()))
  {
  if ( p && (pathname[p-1] !='/') )
    {
    while (p >= 0)
      {
      if (pathname[p] == '/')
        {
        filename = pathname.substr(p+1);
        break;
        }
      --p;
      }
    }
  }
return filename; 
}

void Backup::new_backup(string *conf_name)
{
string dest;
string msg;

update = FALSE;      // assume it is a new backup

if (!fix_conf(conf_name))
  {
  // cerr << "Backup::new_backup - fix_conf FAILED\n";
  return;
  }

// cout << "Backup::new_backup - pathname= \"" << pathname << "\"" << endl;

compression = conf->compression;
if (compression == "GZ")
  {
  dest = pathname + "/gz";
  }
else if (compression == "BZ2")
  {
  dest = pathname + "/bz2";
  }
else if (compression == "GZ_BZ2")
  {
  dest = pathname + "/gz_bz2";
  }
else dest = "";

if (dest.length())
  {
  fstream ofp;
  ofp.open(dest.c_str(), ios::out | ios::binary | ios::trunc );
  ofp.close();
  if (!isfile((char *)dest.c_str()))
    {
    msg = "Can't create \"";
    msg += dest;
    msg += "\" : ";
    msg += strerror(errno);
    msg += " : Quitting";
    m->process_fatal_error(msg);
    return;
    }
  }
process_source_list();
}


// Create the list, including orphans if required
// Used in "update" but not "new backup"
void Backup::get_file_data(string &pathname, list<File_Data> &data, bool orphans, SlaveMessages *m)
{
fstream ifp;
string line;
string fname = pathname + "/data";
File_Data fdata;

data.clear();

ifp.open(fname.c_str(), ios::in | ios::binary);
if (ifp.is_open())
  {
  string str;
  int n;   // position in line
  int len; // length of line

  // We don't know how long a line may be...
  while (getline(ifp, line))
    {
    n = 0;
    len = line.length();

    //cout << "Backup::get_file_data - line= " << line << endl;

    // Remove leading junk
    while  ( isblank(line[n]) && n<len ) ++n;

    if ( (orphans) && (STRIEQN( (char *)line.c_str(), "ORPHAN", 6)) )
      {
      n += 6;
      // Remove leading junk again
      while ( isblank(line[n]) && n<len ) ++n;
      }

    if (line[n++] != '"') continue;         // Probably a comment

    // We have a line
    // ==============

    // Initialize fdata
    fdata.name = "";
    fdata.uid = fdata.gid = fdata.mode = fdata.atime = fdata.mtime = 0;

    while ( (line[n] != '"') && (n<len) )   // get name
      {
      fdata.name += line[n++];
      }

    //cout << "Backup::get_file_data - \"" <<fdata.name << "\"";

    if ( (line[n]=='"') && n<len ) ++n;

    while ( isblank(line[n]) && n<len ) ++n;
    str = "";
    while ( isxdigit(line[n]) && n<len )      // uid
      {
      str += line[n++];
      }
    fdata.uid = hex2ui((char *)str.c_str());

    // cout << " " << str << " " << hex2ui((char *)str.c_str());

    while ( isblank(line[n]) && n<len ) ++n;
    str = "";
    while ( isxdigit(line[n]) && n<len )      // gid
      {
        str += line[n++];
      }
    fdata.gid = hex2ui((char *)str.c_str());        


    //cout << " " << str << " " << hex2ui((char *)str.c_str());

        
    while ( isblank(line[n]) && n<len ) ++n;
    str = "";
    while ( isxdigit(line[n]) && n<len )      // mode
      {
      str += line[n++];
      }
    fdata.mode = hex2ui((char *)str.c_str());

    //cout << " " << str << " " << hex2ui((char *)str.c_str());

    while ( isblank(line[n]) && n<len ) ++n;
    str = "";
    while ( isxdigit(line[n]) && n<len )      // atime
      {
      str += line[n++];
      }
    fdata.atime = hex2ui((char *)str.c_str());

    //cout << " " << str << hex2ui((char *)str.c_str());

    while ( isblank(line[n]) && n<len ) ++n;
    str = "";
    while ( isxdigit(line[n]) && n<len )      // mtime
      {
      str += line[n++];
      }
    fdata.mtime = hex2ui((char *)str.c_str());

    //cout << " " << str << " " << hex2ui((char *)str.c_str()) << endl;;

    // Now add it to the list
    data.push_back(fdata);
    }

  // close the file here....
  ifp.close();

  }
else
  {
  // ERROR, file not opened
  string msg = "Failed to open file \"" + fname;
  msg += "\" for read : ";
  msg += strerror(errno);
  m->process_error(msg);
  }
}


// Incremental backup, updating latest previous one
// OR a new backup if no old backup
// Not sure yet what to do about orphan files
void Backup::update_backup(string *conf_name)
{
string dest;

update = TRUE;

if (!fix_conf(conf_name))
  {
  if (pathname == "")
    {
    new_backup(&pathname);  // Try a new backup instead
    return;
    }
  else return;
  }

// cout << "Backup::update_backup - pathname= " << pathname << endl;

compression = "";
dest = pathname + "/gz";
if (isfile((char *)dest.c_str()))
  {
  compression = "GZ";
  }
else
  {
  dest = pathname + "/bz2";
  if (isfile((char *)dest.c_str()))
    {
    compression = "BZ2";
    }
  else
    {
    dest = pathname + "/gz_bz2";
    if (isfile((char *)dest.c_str()))
      {
      compression = "GZ_BZ2";
      }
        // else compression = "NONE" // the default
    }
  }

//cout << "Backup::update_backup - compression is " << compression << endl;

process_source_list();

}
