/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <iostream>
#include <fstream>
#include <list>
#include <errno.h>

#include "conf.h" 
#include "../common/misc.h"
#include "../common/tfile.h"
#include "../common/version.h"

using namespace std;
using namespace misc;


void Conf::set_defaults(void)
{
// default - keep one old backup
keep_number = 1;

// default - remove files of same name when restoring
//         - if TRUE, those files are saved as pathname_XXX.spiffasave
save_existing_files = FALSE;

// default - if incremental backup detects deleted files these are not restored during a restore
//         - if TRUE then all files are restored
restore_orphans = FALSE;

mount_point = "";
backup_destination = "";
compression = "";

// default pathnames
sh_pname = "/bin/sh";
cat_pname = "/bin/cat";
gzip_pname = "/bin/gzip";
bzip2_pname = "/usr/bin/bzip2";
ln_pname = "/bin/ln";
mtab_pname = "/etc/mtab";
browser_pname = "htmlview %s";

log_type = "APPEND";

conf_pathname = SYSCONFDIR "/spiffa.conf";  // default configuration file
}

Conf::Conf(void)
{
set_defaults();

//cout << "Conf:\n";
load();
}

Conf::Conf(string &pathname)
{
//cout << "Conf(" << pathname << ")\n";
set_defaults();
load(pathname);
}

Conf::~Conf(void)
{
//cout << "~Conf:\n";
}

// Convert a string to upper case
string &Conf::string_to_upper(string &str)
{
string::iterator i;

for (i=str.begin(); i!=str.end(); ++i)
  {
  *i = toupper(*i);
  }
return str;
}

bool Conf::user_permitted(char *user)
{
//  if ( (STREQ(user, "root")) || (glist_index(users, user)!=-1) ) return TRUE;
if ( (STREQ(user, "root")) ||
       (find(users.begin(), users.end(), user) != users.end()))
  {
  return TRUE;
  }
return FALSE;
}

bool Conf::IsSeparator(char x)
{
return ((x == ' ')||(x == '\t')||(x == '=')||(x == ':'));
}

// This should allow quoted strings...
void Conf::getright(string &s, string &v)
{
unsigned int p=0;
bool inquote = FALSE;

v = "";  // start with an empty string!

while ( (p < s.length()) && isblank(s[p]) )
  {
  p++;
  }

if ( (p < s.length()) && ((s[p]=='=') || (s[p]==':')) )
  {
  p++;
  while ( (p < s.length()) && isblank(s[p]) )
    {
    p++;
    }
  }

while (p < s.length())
  {
  if (s[p] == '\"')
    {
    inquote = !inquote;
    p++;
    continue;
    }

  // If any of these the rest of the line is ignored unless in quote
  if ( (!inquote) &&
    (isblank(s[p]) || (s[p]=='#') || (s[p]==';') || (s[p]=='\n')) ) break;

    v += s[p];
    p++;
  }
}


// Load configuration
void Conf::load(string &pathname)
{
fstream ifp;

error_str = "";

if (!isfile((char *)pathname.c_str()))
  {
  error_str = "ERROR: Pathname \"" + pathname;
  error_str += "\" is not a file: can't read it";

  cerr << "Conf::load - " << error_str << endl;

  return;
  }

ifp.open(pathname.c_str(), ios::in | ios::binary );

if ( ifp.is_open() )
  {
  string line;
  string r;
  string s;
  int num = 0;

  while ( getline(ifp, line) )
    {
    while (isblank(line[0])) line = line.substr(1);
    if (line.length() == 0) continue;

    if ((num == 0) && (line != "# Configuration file for spiffa"))
      {
      error_str = "ERROR: \"" + pathname;
      error_str += "\" ignored, invalid configuration file.";
      cerr << error_str << endl;
      ifp.close();
      return;
      }
    else num++;

    if ( (line[0] == '#') || (line[0] == ';') || (line.length() == 0) )
      {
      continue; // comment line or empty
      }

    if (STRIEQN("BACKUP_SOURCE", (char *)line.c_str(), 13) && IsSeparator(line[13]) )
      {
      list<string>::iterator i;
      bool found = FALSE;
 
      s = line.substr(13);
      getright(s, r);
      remove_leading_and_terminal_space(r);
      if (r.length())
        {
        for (i=backup_sources.begin(); i!=backup_sources.end(); ++i)
          {
          if ((*i)==r)
            {
            found = TRUE;   // Already in the list
            break;
            }

          if ( r == i->substr(0, r.length()))
            {
            found = TRUE;   // Included by existing entry
            break;
            }

          if ( (*i) == r.substr(0, i->length()))
            {
            // Remove existing entry and put in this one, this includes existing entry
            backup_sources.erase(i);
            break;
            }
          }

        if (!found)
          {
          backup_sources.push_back(r);
          }
        }
      }
    else if (STRIEQN("BACKUP_DENY", (char *)line.c_str(), 11) && IsSeparator(line[11]) )
      {
      list<string>::iterator i;
      bool found = FALSE;

      s = line.substr(11);
      getright(s, r);
      remove_leading_and_terminal_space(r);
      if (r.length())
        {
        for (i=backup_deny.begin(); i!=backup_deny.end(); ++i)
          {
          if ((*i)==r)
            {
            found = TRUE;   // Already in the list
            break;
            }

          if ( r == i->substr(0, r.length()))
            {
            found = TRUE;   // Included by existing entry
            break;
            }

          if ( (*i) == r.substr(0, i->length()))
            {
            // Remove existing entry and put in this one, this includes existing entry
            backup_deny.erase(i);
            break;
            }
          }

        if (!found)
          {
          backup_deny.push_back(r);
          }
        }
      }
    else if (STRIEQN("USER", (char *)line.c_str(), 4)  && IsSeparator(line[4]) )
      {
      s = line.substr(4);
      getright(s, r);
      remove_leading_and_terminal_space(r);
      if (r.length())
        {
        if (find(users.begin(), users.end(), r) == users.end())
          {
          users.push_back(r);
          }
        }
      }
    else if (STRIEQN("MOUNT_POINT", (char *)line.c_str(), 11) && IsSeparator(line[11]) )
      {
      s = line.substr(11);
      getright(s, r);
      remove_leading_and_terminal_space(r);
      if (r.length())
        {
        mount_point = r;
        }
      }
    else if (STRIEQN("BACKUP_DESTINATION", (char *)line.c_str(), 18) && IsSeparator(line[18]) )
      {
      s = line.substr(18);
      getright(s, r);
      remove_leading_and_terminal_space(r);
      if (r.length())
        {
        backup_destination = r;
        }
      }
    else if (STRIEQN("COMPRESSION", (char *)line.c_str(), 11) && IsSeparator(line[11]) )
      {
      s = line.substr(11);
      getright(s, r);
      remove_leading_and_terminal_space(r);
      if (r.length())
        {
        compression = string_to_upper(r);
        }
      }
    else if (STRIEQN("KEEP_NUMBER", (char *)line.c_str(), 11) && IsSeparator(line[11]) )
      {
      s = line.substr(11);
      getright(s, r);
      keep_number = atoi(r.c_str());
      }
    else if (STRIEQN("SAVE_EXISTING_FILES", (char *)line.c_str(), 19) && IsSeparator(line[19]) )
      {
      s = line.substr(19);
      getright(s, r);
      string_to_upper(r);
      if ( (r=="YES") || (r=="TRUE") || (r=="1") )
        {
        save_existing_files = TRUE;
        }
      }
    else if (STRIEQN("RESTORE_ORPHANS", (char *)line.c_str(), 15) && IsSeparator(line[15]) )
      {
      s = line.substr(15);
      getright(s, r);
      string_to_upper(r);
      if ( (r=="YES") || (r=="TRUE") || (r=="1") )
        {
        restore_orphans = TRUE;
        }
      }
    else if (STRIEQN("SH_PATHNAME", (char *)line.c_str(), 11) && IsSeparator(line[11]) )
      {
      s = line.substr(11);
      getright(s, r);
      remove_leading_and_terminal_space(r);
      if (r.length())
        {
        sh_pname = r;
        }
      }
    else if (STRIEQN("CAT_PATHNAME", (char *)line.c_str(), 12) && IsSeparator(line[12]) )
      {
      s = line.substr(12);
      getright(s, r);
      remove_leading_and_terminal_space(r);
      if (r.length())
        {
        cat_pname = r;
        }
      }
    else if (STRIEQN("GZIP_PATHNAME", (char *)line.c_str(), 13) && IsSeparator(line[13]) )
      {
      s = line.substr(13);
      getright(s, r);
      remove_leading_and_terminal_space(r);
      if (r.length())
        {
        gzip_pname = r;
        }
      }
    else if (STRIEQN("BZIP2_PATHNAME", (char *)line.c_str(), 14) && IsSeparator(line[14]) )
      {
      s = line.substr(14);
      getright(s, r);
      remove_leading_and_terminal_space(r);
      if (r.length())
        {
        bzip2_pname = r;
        }
      }
    else if (STRIEQN("LN_PATHNAME", (char *)line.c_str(), 11) && IsSeparator(line[11]) )
      {
      s = line.substr(11);
      getright(s, r);
      remove_leading_and_terminal_space(r);
      if (r.length())
        {
        ln_pname = r;
        }
      }
    else if (STRIEQN("MTAB_PATHNAME", (char *)line.c_str(), 13) && IsSeparator(line[13]) )
      {
      s = line.substr(13);
      getright(s, r);
      remove_leading_and_terminal_space(r);
      if (r.length())
        {
        mtab_pname = r;
        }
      }
    else if (STRIEQN("BROWSER_PATHNAME", (char *)line.c_str(), 16) && IsSeparator(line[16]) )
      {
      s = line.substr(16);
      getright(s, r);
      remove_leading_and_terminal_space(r);
      if (r.length())
        {
        browser_pname = r;
        }
      }
    else if (STRIEQN("LOG_TYPE", (char *)line.c_str(), 8) && IsSeparator(line[8]) )
      {
      s = line.substr(8);
      getright(s, r);
      remove_leading_and_terminal_space(r);
      if (r.length())
        {
        log_type =  string_to_upper(r);
        }
      }
    else
      {
      // UNKNOWN, ignore
      }
    }

  if (find(backup_deny.begin(), backup_deny.end(), "/proc") == backup_deny.end())
    {
     backup_deny.push_back("/proc");
    }

  if (find(backup_deny.begin(), backup_deny.end(), "/sys") == backup_deny.end())
    {
     backup_deny.push_back("/sys");
    }

  ifp.close();
  }

else
  {
  error_str = "ERROR: Can't open \"" + pathname;
  error_str += "\" for read : ";
  error_str += strerror(errno);
  cerr << "Conf::load - " << error_str << endl;
  }

}

// Load default configuration
void Conf::load(void)
{
load(conf_pathname);
}


// Save configuration
void Conf::save(string &pathname)
{
fstream ofp;

error_str = "";
// Overwrite any existing file
ofp.open(pathname.c_str(), ios::out | ios::binary | ios::trunc );
if ( ofp.is_open() )
  {
  list<string>::iterator i;

  ofp << "# Configuration file for spiffa\n\n";
  
  ofp << "# Source files and directories for backup\n";
  for (i=backup_sources.begin(); i!=backup_sources.end(); ++i)
    {
    ofp << "BACKUP_SOURCE = " << "\"" << *i << "\"" << endl;
    }

  ofp << "\n# Source files and directories NOT for backup\n";
  ofp << "# If the contents of a directory are to be omitted\n";
  ofp << "# use a trailing slash with the directory name.\n";
  ofp << "# The directory name will then be included but not\n";
  ofp << "# its contents.  Without the trailing slash the directory\n";
  ofp << "# name and its contents are omitted.\n";
  for (i=backup_deny.begin(); i!=backup_deny.end(); ++i)
    {
    ofp << "BACKUP_DENY = " << "\"" << *i << "\"" << endl;
    }

  ofp << "\n# Only the following ordinary users are permitted to use spiffa\n";
  ofp << "# The superuser \"root\" doesn't need to be in this list\n";  
  for (i=users.begin(); i!=users.end(); ++i)
    {
    ofp << "USER = " << *i << endl;
    }

  ofp << "\n# If the destination directory for the backup\n";
  ofp << "# is mounted on a permanent mount point that is present\n";
  ofp << "# even when it is not mounted, such as used with \"/etc/fstab\",\n";
  ofp << "# specify it here so that spiffa won't write to it unless mounted.\n";
  ofp << "MOUNT_POINT = " << "\"" << mount_point << "\"" << endl;

  ofp << "\n# The destination directory for the backup\n";
  ofp << "# This is also the source directory for a restore\n";
  ofp << "BACKUP_DESTINATION = " << "\"" << backup_destination << "\"" << endl;
  
  ofp << "\n# The type of compression to use\n";
  ofp << "# This can be GZ BZ2 GZ_BZ2 or NONE\n";
  ofp << "COMPRESSION = " << compression << endl;

  ofp << "\n# The number of old backups to retain in addition to the current one.\n";
  ofp << "# The value should be from -1 to 100.\n";
  ofp << "# If this is -1 then all old backups are retained.\n";
  ofp << "KEEP_NUMBER = " << TFile::itoa(keep_number) << endl;

  ofp << "\n# Save copies of old files\n";
  ofp << "# If this is YES or TRUE or 1 then files that would be overwritten\n";
  ofp << "# during a restore are retained with XXX.spiffasave as a suffix\n";
  ofp << "# where XXX is a decimal number.\n";
  ofp << "SAVE_EXISTING_FILES = " << ( (save_existing_files) ? "YES" : "NO" ) << endl;

  ofp << "\n# Restore orphan files\n";
  ofp << "# When an incremental backup is performed,\n";
  ofp << "# files that have been deleted in the source since the initial backup\n";
  ofp << "# are marked as orphans.  If this is YES or TRUE or 1 then these files\n";
  ofp << "# will be restored in addition to the rest, otherwise they are ignored.\n";
  ofp << "RESTORE_ORPHANS = " << ( (restore_orphans) ? "YES" : "NO" ) << endl;

  ofp << "\n# Pathnames of commands etc. used by spiffa and/or slave\n";
  ofp << "SH_PATHNAME = \""      << sh_pname      << "\"" << endl;
  ofp << "CAT_PATHNAME = \""     << cat_pname     << "\"" << endl;
  ofp << "GZIP_PATHNAME = \""    << gzip_pname    << "\"" << endl;
  ofp << "BZIP2_PATHNAME = \""   << bzip2_pname   << "\"" << endl;
  ofp << "LN_PATHNAME = \""      << ln_pname      << "\"" << endl;
  ofp << "MTAB_PATHNAME = \""    << mtab_pname    << "\"" << endl;
  ofp << "BROWSER_PATHNAME = \"" << browser_pname << "\"" << endl;

  ofp << "\n# Options for spiffa.log\n";
  ofp << "# This can be NEW, APPEND or NONE\n";
  ofp << "# The file \"" SYSCONFDIR"/spiffa.log\" can be opened as a new file\n";
  ofp << "# or it can be appended to the existing file if present.  Alternatively\n";
  ofp << "# it may be preferred to have no file at all.\n";
  ofp << "LOG_TYPE = " << log_type << endl;

  ofp.close();

  // ONLY ROOT should be able to read/write.
  chown(pathname.c_str(), 0, 0);  // owned by root, group root
  chmod(pathname.c_str(), 0660);  // read/write by root, group root

  // Check that mount point aggrees with destination
  if ( mount_point.length() && 
        (!STREQN((char *)backup_destination.c_str(),
            (char *)mount_point.c_str(),
            mount_point.length())) )
    {
    error_str = "WARNING: Mount Point and destination DON'T AGREE!";
    }
  }
else
  {
  error_str = "ERROR: Can't open \"" + pathname;
  error_str += "\" for write : ";
  error_str += strerror(errno);

  cerr << "Conf::save - " << error_str << endl;
  }
}

// Save default configuration
void Conf::save(void)
{
  save(conf_pathname);
}

