/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#include <iostream>
#include <fstream>
#include <dirent.h>
#include <sys/types.h>
#include <pwd.h>
#include <ctype.h>
#include <utime.h>
#include <errno.h>
#include <stdlib.h>
#include <linux/limits.h>  // PATH_MAX and NAME_MAX

#include "restore.h"
#include "conf.h"
#include "backup.h"
#include "../common/misc.h"
#include "../common/tfile.h"
#include "new_process.h"
#include "../common/messages.h"
#include "../common/version.h"

using namespace std;
using namespace misc;


Restore::~Restore()
{
  // cout << "~Restore:\n";

  data.clear();

  if (conf)
    {
      delete conf;
      conf = NULL;
    }
  if (errors_ofp.is_open())
    {
      errors_ofp.close();
    }
}


bool Restore::set_dest_attributes(string *dest, bool link)
{
  string msg;

  if (lchown(dest->c_str(), 0, 0))
    {
      // ERROR: Can't chown!  - possibly not important - IT IS IMPORTANT!
      msg = "Can't change ownership of \"" + *dest;
      msg += "\" to root : ";
      msg += strerror(errno);
      m->process_error(msg);

      // cout << "Restore::set_attributes - lchown FAILED\n";
      return FALSE;
    }

  if ((!link) && (chmod(dest->c_str(), 0644)))
    {
      // ERROR: Can't chmod!  - possibly not important - IT IS IMPORTANT!
      msg = "Can't change permissions of \"" + *dest;
      msg += "\" to 644 : ";
      msg += strerror(errno);
      m->process_error(msg);

      // cout << "Restore::set_attributes - chmod FAILED\n";
      return FALSE;
    }
  return TRUE;
}


bool Restore::create_directory(string *name)
{
  string msg;

  if (mkdir(name->c_str(), 0644))
    {
      // ERROR: Can't create!
      msg = "Failed to create directory \"";
      msg += *name;
      msg += "\" : ";
      msg += strerror(errno);

      m->process_error(msg);
      return FALSE;
    }

  return (set_dest_attributes(name, FALSE));

}

bool Restore::find_backup(string *pathname, Conf *c, SlaveMessages *m )
{
  DIR *dp;
  struct dirent *de;
  string s_name;
  string msg;
    
  double stamp = 0;
  double test_stamp;
  string hostname;
  char buffer[128]; // unnecessarily large!
  char *b;          // position in buffer
  char *p;          // position in current name being checked
  string basename;

  // cout << "Restore::find_backup\n";

  find_hostname(&hostname, &msg);
  if (msg.length())
    {
    m->process_warning(msg);
    }
    
  basename = "spiffa_backup_";
  basename += hostname;
  basename += "_";

  dp = opendir(c->backup_destination.c_str());
  if (dp)
    {
      while ( (de = readdir(dp)) )
  {
    //cout <<  "Restore::find_backup - got1 " << de->d_name << endl;

    if ( STREQN(de->d_name, (char *)basename.c_str(), basename.length() ) )
      {
     
        // Get its datestamp from its name (YYYY_MM_DD_hh_mm_ss)
        p = &de->d_name[basename.length()]; // skip over the basename

        b = buffer;
 
        if (strlen(p) != 19)
    {
      // BAD NAME
      // cout <<  "Restore::find_backup - BAD NAME1 " << p << endl;
      continue;
    }
        while (*p)
    {
      if (*p == '_') {;} // skip a '_'
      else if ( (*p) && isdigit(*p) ) *b++ = *p;
      else
        {
          // BAD NAME
          // cout <<  "Restore::find_backup - BAD NAME2 " << p << endl;
          break;
        }
      p++;
    }
        *b = 0;
            
        if ((strlen(buffer) != 14) || *p)
    {
      // BAD NAME
      // cout <<  "Restore::find_backup - BAD NAME3 " << p << endl;
      continue;
    }

        test_stamp = atof(buffer);
          
        if (test_stamp > stamp)
    {
      stamp = test_stamp;
      s_name = c->backup_destination;
      if (c->backup_destination[c->backup_destination.length() - 1] != '/')
        {
          s_name += '/';
        }
      s_name += de->d_name;
    }
      }
  } // End of while readdir
      closedir(dp);
    }
  if (s_name.length())
    {
      *pathname = s_name;
      return TRUE;
    }
  else *pathname = "";
  return FALSE;
}


// Tthe parent pathname of a "from" file is needed
// so that later the compression type etc. can be obtained
bool Restore::get_parent_pathname(string *from, string *pathname)
{
  string basename;
  string hostname;
  string name;
  unsigned int n;
  string msg;
  
  find_hostname(&hostname, &msg);
  if (msg.length())
    {
    m->process_warning(msg);
    }

  basename = "spiffa_backup_";
  basename += hostname;
  basename += "_";

  for (n=0; n<from->length(); ++n)
    {
      if ( STREQN( (char *)(from->substr(n)).c_str(),
       (char *)basename.c_str(),
       basename.length() ) )
        {
        unsigned int ln = 0;

        string directory = from->substr(0, n);

        while ( (n + ln) <= from->length() )
            {
            if ( ( (*from)[n+ln] == '/' ) || ( (n+ln) == from->length() ) )
                {
                *pathname = directory + name;
                return ( ((ln - basename.length()) == 19) ? TRUE : FALSE );
                }
            name += (*from)[n+ln];
            ln++;
            }
        break;
        }
    }

  return FALSE;
}


// Return TRUE if the pathname is in the data list

bool Restore::name_in_list(string &pathname, list<File_Data>::iterator &p)
{
list<File_Data>::iterator i;

for (i=data.begin(); i!=data.end(); ++i)
    {
    if ( (*i).name == pathname)
        {
        p = i;
        return TRUE;
        }
    }
p = i;
return FALSE;
}



// Return TRUE if dest and src are identical
// The permissions have to be suitable for the check!
// dest has just been written, but src could be a problem

bool Restore::files_identical(string *dest, string *src)
{
  struct stat64 statbuf;
  string msg;

  if (Backup::get_statbuf(src, &statbuf, m))
    {
      off64_t size;
      fstream s_fp;
      fstream d_fp;
      char s, d;
      size_t n;

      if (! get_fsize((char *)dest->c_str(), &size))
  {
    // ERROR
   
    msg = "Can't get size of \"";
    msg += *dest;
    msg += "\": ";
    msg += strerror(errno);
    m->process_error(msg);
    return(FALSE);
  }

      // Now make sure we can read it and the sizes are the same
      if (!Backup::change_permissions_to_0777(src, m))
  {
    Backup::restore_permissions(src, &statbuf, m);
    return(FALSE);
  }



      // - make sure the sizes are the same
      if (size != statbuf.st_size)
  {
    Backup::restore_permissions(src, &statbuf, m);
    return  FALSE;  // not the same
  }

      // Open source
      s_fp.open(src->c_str(), ios::in  | ios::binary );
      if (!s_fp.is_open())
  {
    // ERROR
    msg = "Can't open file \"";
    msg += *src;
    msg += "\" for read : ";
    msg += strerror(errno);
    m->process_error(msg);

    Backup::restore_permissions(src, &statbuf, m);
    return  FALSE;  // not the same
  }

      // Open destination
      d_fp.open(dest->c_str(), ios::in  | ios::binary );
      if (!d_fp.is_open())
  {
    // ERROR
    msg = "Can't open file \"";
    msg += *dest;
    msg += "\" for read : ";
    msg += strerror(errno);
    m->process_error(msg);

    Backup::restore_permissions(src, &statbuf, m);
    return FALSE;  //  NOT THE SAME
  }

      n = 0;
      // Now perform the compare
      while ((s_fp.get(s)) && d_fp.get(d))
  {
    if (s != d)
      {
        // NOT THE SAME
        s_fp.close();
        d_fp.close();

        Backup::restore_permissions(src, &statbuf, m);
        return FALSE;  //  NOT THE SAME
      }
    n++;
  }

      s_fp.close();
      d_fp.close();

      if (n != size)
  {
    // NOT THE SAME

    Backup::restore_permissions(src, &statbuf, m);
    return FALSE;  //  NOT THE SAME
  }

      Backup::restore_permissions(src, &statbuf, m);
      return TRUE; 
    }
  else return FALSE;
}

// FILE
//
// NOTE: Only regular files are possibly compressed
//       All other types keep their original form
//
// If the destination file already exists, rename it
// When restored, delete the renamed file unless we are saving existing files
// If restore fails, then rename it back to the original
bool Restore::process_file(string *dest, string *s_file)
{
string command;
string temp;
bool file_exists = FALSE;
string msg;
list<File_Data>::iterator p;
string s;
string d;

if (!(name_in_list(*dest, p)))
    {
      msg = "Couldn't find file \"" + *dest;
      msg += "\" in the list of file data";
      m->process_error(msg);
      return FALSE;
   }


  // If the "dest" file exists already, create a temp filename
  if (isfile((char *)dest->c_str()))
    {
      // This file already exists so we need to rename it as a saved file for now

      TFile tf;
      string suffix = ".spiffasave";
      string name = *dest + "_";
      temp = *tf.tname(&name, 3, &suffix);
      if (rename(dest->c_str(), temp.c_str()))
  {
    // ERROR - couldn't rename
    msg = "Failed rename pre-existing destination file \"";
    msg += *dest;
    msg += "\" as \"";
    msg += temp;
    msg += "\" : ";
    msg += strerror(errno);
    m->process_error(msg);
    return FALSE;
  }
      else file_exists = TRUE;
    }

  Backup::fix_name(s, s_file);
  Backup::fix_name(d, dest);

  if (compression == "gz")
    {    
      command = conf->cat_pname + " '";
      command += s;
      command += "' | ";
      command += conf->gzip_pname;
      command += " -d -N > '";
      command += d;
      command += "'";
    }
  else if (compression == "bz2")
    {  
      command = conf->cat_pname + " '";
      command += s;
      command += "' | ";
      command += conf->bzip2_pname;
      command += " -d > '";
      command += d;
      command += "'";
    }
  else if (compression == "gz_bz2")
    {
      command = conf->cat_pname + " '";
      command += s;
      command += "' | ";
      command += conf->bzip2_pname;
      command += " -d | ";
      command += conf->gzip_pname;
      command += " -d -N > '";
      command += d;
      command += "'";
    }
  else  // DEFAULT - assume that (Conf::compression == "NONE")  
    {
      command = conf->cat_pname + " '";
      command += s;
      command += "' > '";
      command += d;
      command += "'";
    }

  if (run.new_process(command, conf->sh_pname)) 
    {
    // ERROR
    msg = "Command \"";
    msg += command;
    msg += "\" FAILED : ";
    msg += run.status_string;
    m->process_error(msg);

    if (file_exists)
  {
    // Rename the saved file back to its original name
    if (rename(temp.c_str(), dest->c_str()))
      {
        // ERROR - couldn't rename
        msg = "Failed to rename temporary file \"";
        msg += temp;
        msg += "\" back to \"";
        msg += *dest;
        msg += "\" : ";
        msg += strerror(errno); 
        m->process_error(msg);
      }
  }

      return FALSE;
    }



  // cout << "Restore::process_file - dest= " << *dest << endl;
  // cout << "Restore::process_file -  src= " << *s_file << endl;

// Check that the destination was created! 
if (!isfile((char *)dest->c_str()))
  {
    // ERROR
    // cout << "Restore::process_file - FAILED to restore " << dest << endl;

    if (file_exists)
      {
  // Rename the saved file back to its original name
  if (rename(temp.c_str(), dest->c_str()))
    {
      // ERROR - couldn't rename
      msg = "Failed to rename temporary file \"";
      msg += temp;
      msg += "\" back to \"";
      msg += *dest;
      msg += "\" : ";
      msg += strerror(errno); 
      m->process_error(msg);
    }
      }
    return FALSE;
  }


// This destroys the old file unless we want to retain it.
// If the files are identical, the old file should be removed anyway
 else if (file_exists)
  {
    if ( (!conf->save_existing_files) || files_identical(dest, &temp) )
      {
  if (remove(temp.c_str()))
    {
      // ERROR
      msg = "Can't remove file \"";
      msg += temp;
      msg += "\" : ";
      msg += strerror(errno);
      m->process_warning(msg);

      // cout <<  "Restore::process_file - " << msg << endl;
    }
      }
  }

// Success so save the iterator
data_pointer.push_back(p);
m->num_names++;
return TRUE;
}


bool Restore::same_target(string *name, string &target)
{
  string name_target;

  if (Backup::get_target(name, name_target, m))
    {
      if (name_target == target)
  {
    return TRUE;
  }
    }
  return FALSE;
}



// LINK
// set attributes at end

// Should be copied if required...
// =============================== 
bool Restore::process_link(string *d_link, string *s_link)
{
string command;
string temp;
bool link_exists = FALSE;
list<File_Data>::iterator p;
string target;
string msg;

string t;
string d;

if (!(name_in_list(*d_link, p)))
    {
    msg = "Couldn't find link \"" + *d_link;
    msg += "\" in the list of file data";
    m->process_error(msg);
    return FALSE;
    }

  if (!Backup::get_target(s_link, target, m)) return FALSE;

  Backup::fix_name(t, &target);
  Backup::fix_name(d, d_link);

 // If the link exists already, create a temp filename
 if (islink((char *)d_link->c_str()))
   {
     // This link already exists so we need to rename it as a saved file for now

     TFile tf;
     string suffix = ".spiffasave";
     string name = *d_link + "_";
     temp = *tf.tname(&name, 3, &suffix);
     if (rename(d_link->c_str(), temp.c_str()))
       {
   // ERROR - couldn't rename
   msg = "Failed to rename pre-existing destination link \"";
   msg +=  *d_link;
   msg += "\" as \"";
   msg += temp;
   msg += "\" : ";
   msg += strerror(errno);
   m->process_error(msg);
   return FALSE;
       }
     else link_exists = TRUE;
   }


 command = conf->ln_pname + " -s \'";
 command += t;
 command += "\' \'";
 command += d;           // the link name that is created
 command += "\'";

 if (run.new_process(command, conf->sh_pname)) 
   {
    // ERROR
    msg = "Command \"";
    msg += command;
    msg += "\" FAILED : ";
    msg += run.status_string;
    m->process_error(msg);
    return FALSE;
   }

 if (!islink((char *)d_link->c_str()))
   {
     // ERROR
     msg = "Failed to create destination link \"";
     msg += *d_link;
     msg += "\" although command to do so succeded! : ";
     msg += strerror(errno);
     m->process_error(msg);

     if (link_exists)
       {
   // Rename the saved file back to its original name
   if (rename(temp.c_str(), d_link->c_str()))
     {
       // ERROR
       msg = "Failed to rename temporary link \"";
       msg += temp;
       msg += "\" back to \"";
       msg += *d_link;
       msg += "\" : ";
       msg += strerror(errno);
       m->process_error(msg);
     }
       }
     return FALSE;
   }

 // For this we need to be able to check if the link is different
 // Does it have a different target?

 else if (link_exists)
   {
    if ( (!conf->save_existing_files) || same_target(&temp, target) )
      {
  if (remove(temp.c_str()))
    {
      // ERROR
      msg = "Failed to delete temporary link \"";
      msg += temp;
      msg += "\" : ";
      msg += strerror(errno);
      m->process_warning(msg);
    }
      }
   }
   
// Success so save the iterator

data_pointer.push_back(p);
m->num_names++;
return TRUE;
}






// ======================================================================
// The code for devices, fifos and sockets is almost the same
// Only the mknode function is different, so pass an int 0, 1 or 2
// for device, fifo or socket respectively
// then do a switch
// ======================================================================
//bool Restore::process_node(string *d_node, string *s_node, int type)
bool Restore::process_node(string *d_node, string *s_node)
{
struct stat64 statbuf;
struct stat64 d_statbuf;
int d_type = -1;
int s_type = -1;
string temp;
bool node_exists = FALSE;
list<File_Data>::iterator p;
string msg;
int rc;

  if (!(name_in_list(*d_node, p)))
    {
      // ERROR
      msg = "Couldn't find special file \"" + *d_node;
      msg += "\" in the list of file data";
      m->process_error(msg);
      return FALSE;
    }

  // If the node already exists
  if (Backup::get_statbuf(d_node, &d_statbuf, m))
    {
      // RENAME THE EXISTING FILE

      TFile tf;
      string suffix = ".spiffasave";
      string name = *d_node + "_";

      temp = *tf.tname(&name, 3, &suffix);
      if (rename(d_node->c_str(), temp.c_str()))
  {
    // ERROR - couldn't rename
    msg = "Failed to rename pre-existing destination special file \"";
    msg +=  *d_node;
    msg += "\" as \"";
    msg += temp;
    msg += "\" : ";
    msg += strerror(errno);
    m->process_error(msg);
    return FALSE;
  }

      if (S_ISCHR(d_statbuf.st_mode))
  d_type = PROCESS_CHAR_DEVICE;

      else if (S_ISBLK(d_statbuf.st_mode))
  d_type = PROCESS_BLOCK_DEVICE;

      else if (S_ISFIFO(d_statbuf.st_mode))
  d_type = PROCESS_FIFO;

      else if (S_ISSOCK(d_statbuf.st_mode))
  d_type = PROCESS_SOCKET;

      node_exists = TRUE;
    }

  if (!(Backup::get_statbuf(s_node, &statbuf, m)))
    {
      return FALSE;
    }


  if ( (S_ISCHR(statbuf.st_mode)) || (S_ISBLK(statbuf.st_mode)) )
    {
      s_type = (S_ISCHR(statbuf.st_mode)) ? PROCESS_CHAR_DEVICE : PROCESS_BLOCK_DEVICE;
      rc = mknod(d_node->c_str(), statbuf.st_mode, statbuf.st_rdev);
    }
  else if (S_ISFIFO(statbuf.st_mode))
    {
      s_type = PROCESS_FIFO;
      rc = mkfifo(d_node->c_str(), statbuf.st_mode);
    }
  else if (S_ISSOCK(statbuf.st_mode))
    {
      s_type = PROCESS_SOCKET;
      rc = mknod(d_node->c_str(), statbuf.st_mode, 0);
    }
  else rc = -2;

  if (rc)
    {
      if (rc == -2) msg = "UNKNOWN TYPE OF SPECIAL FILE \"";
      else msg = "Failed to process special file \"";
      msg += *s_node;
      msg += "\" : ";
      msg += strerror(errno);

      if (node_exists)
  {
    if (rename(temp.c_str(), d_node->c_str()))
      {
        // ERROR
        msg += " ALSO: Failed to rename temporary special file \"";
        msg += temp;
        msg += "\" back to \"";
        msg += *d_node;
        msg += "\" : ";
        msg += strerror(errno);
      }
  }

      m->process_error(msg);
      return FALSE;
    }

  else if (node_exists)
    {
      if ( (!conf->save_existing_files) || (s_type == d_type) )
  {
    if (remove(temp.c_str()))
      {
        // ERROR
        msg = "Failed to rename temporary special file \"";
        msg += temp;
        msg += "\" back to \"";
        msg += *d_node;
        msg += "\" : ";
        msg += strerror(errno);
        m->process_error(msg);
      }
  }
    }

// Success so save the iterator
data_pointer.push_back(p);
m->num_names++;
return TRUE;
}

bool Restore::process_directory(string *d_dir, string *s_dir)
{
list<File_Data>::iterator p;
DIR *dp;
struct dirent *de;
string s_name;
string d_name;

if (m->quit_flag) return FALSE;

//cout << "\nRestore::process_dir - s directory " << *s_dir << endl;
//cout   << "Restore::process_dir - d directory " << *d_dir << endl;

// root directory / is assumed to exist...
if ( (!(*d_dir == "/")) && (!(name_in_list(*d_dir, p))) )
  {
  // cout << "Restore::process_directory - " << *d_dir << " NOT IN LIST\n";
  return FALSE;
  }

dp = opendir(s_dir->c_str());
if (dp)
  {
  //  cout << "\nRestore::process_dir - opened directory " << *s_dir << "\n";
  
  while ( (de = readdir(dp)) )
    {
    //    cout << "Restore::process_dir - got dirent for " << de->d_name << endl;
    if (m->quit_flag) break;

    if ( STREQ(de->d_name, ".") || STREQ(de->d_name, "..")) continue;
      
    s_name = *s_dir;
    if (s_name.length() && s_name[s_name.length()-1] != '/') s_name += "/";
    s_name += de->d_name;

    d_name = *d_dir;
    if (d_name.length() == 0) d_name = "/";
    else if (d_name[d_name.length()-1] != '/') d_name += "/";
    d_name += de->d_name;
    
    //    cout << "\nRestore::process_dir - de->d_name " << de->d_name << "\n";

    //      cout << "\nRestore::process_dir - d_name " << d_name << "\n";


    process_name(d_name, s_name, de->d_name);
    }
  closedir(dp);
  }

// Success so save the iterator
if (!(*d_dir == "/")) // Assumed to exist...
    {
    data_pointer.push_back(p);
    m->num_names++;
    }
return TRUE;
}

 // Remove the ".gz", ".bz2" or ".gz.bz2" extension from the destination name
 // Converts dest to pathname (and filename to fname if not NULL)
void Restore::remove_compression_suffix(string &pathname, string &dest, string *fname, char *filename)
{
if (compression == "gz")
    {
    if (SUFEQ( (char *)dest.c_str(), ".gz") )
        {
        pathname = dest.substr(0, dest.length() - 3);
        if (fname)
            {
            *fname = "";
            for (unsigned int n=0; n<(strlen(filename)-3); ++n)
                {
                *fname += filename[n];
                }
            }
        }
    }
else if (compression == "bz2")
    {
    if (SUFEQ( (char *)dest.c_str(), ".bz2") )
        {
        pathname = dest.substr(0, dest.length() - 4);
        if (fname)
            {
            *fname = "";
            for (unsigned int n=0; n<(strlen(filename)-4); ++n)
                {
                *fname += filename[n];
                }
            }
        }
    }
else if (compression == "gz_bz2")
    {
    if (SUFEQ( (char *)dest.c_str(), ".gz.bz2") )
        {
        pathname = dest.substr(0, dest.length() - 7);
        if (fname)
            {
            *fname = "";
            for (unsigned int n=0; n<(strlen(filename)-7); ++n)
                {
                *fname += filename[n];
                }
            }
        }
    }
else
    {
    // No Compression
    pathname = dest;
    if (fname) *fname = filename;
    }
}


// May want this to return bool FALSE for failure...
void Restore::process_name(string &dest, string &source, char *filename)
{
string msg;

m->get_messages();
if (m->quit_flag) return;

if (isdir((char *)source.c_str()))
    {
    struct stat64 statbuf;

    m->send_message(SL_PATH, (char *)dest.c_str());

    // cout << "Restore::process_name - source " << source << " is DIRECTORY" << endl;
    if (!(dest=="/")) // root directory is assumed to exist...
        {
        if (!(isdir((char *)dest.c_str())))
            {
            if (!create_directory(&dest))
                {
                return;
                }
            }
        }

    if (!(Backup::get_statbuf(&dest,  &statbuf, m)))
        {
        return;
        }

    // Now make sure we can read it
    Backup::change_permissions_to_0777(&dest, m);

    if (process_directory(&dest, &source))
        {
        Backup::restore_permissions(&dest, &statbuf, m);
        }
    }
else if (isfile((char *)source.c_str()))
    {
    string pathname;
    string fname;

    remove_compression_suffix(pathname, dest, &fname, filename);

    // This needs the destination filename,
    // not the source filename which may be .gz or .bz
    m->send_message(SL_FILE, (char *)fname.c_str());
      
    if (!process_file(&pathname, &source))
        {
        // ERROR
        msg = "FAILED to restore file \"";
        msg += pathname;
        msg += "\"";
        m->process_error(msg);
        }
    }
else // link or special file
    {
    m->send_message(SL_FILE, filename);

    if (islink((char *)source.c_str()))
        {
        // cout << "Restore::process_name - source " << source << " is LINK" << endl;
      
        if (!process_link(&dest, &source))
            {
            // ERROR
            msg = "FAILED to restore link \"";
            msg += dest;
            msg += "\"";
            m->process_error(msg);
            }
        }
    // Anything else is either a node or an error
    else if (!process_node(&dest, &source))
        {
        // ERROR
        msg = "FAILED to restore special file \"";
        msg += dest;
        msg += "\"";
        m->process_error(msg);
        }
    }
}


// NOTE: BSD has a function lutimes() which doesn't follow symbolic links
// ======================================================================

// Fix up uid gid mode atime and mtime for all restored files and directories
// Can't fix creation time though
void Restore::fix_uid_etc(void)
{
list<list<File_Data>::iterator>::iterator p;
struct utimbuf utb;
string msg;

//cout << "Restore::fix_uid_etc" << endl;

p=data_pointer.begin();

//cout << "m->num_names= " << m->num_names << endl;

for (int n=0; n<m->num_names; ++n, ++p)
    {
//    cout << n << " : " << (**p).name << " : " << (**p).backup_name << " : " 
//            << (**p).mode << " : " <<  (**p).atime << " : "  <<  (**p).mtime <<  endl;

    if (chown( (**p).name.c_str(),
                        (**p).uid,
                        (**p).gid ) )
        {
        // ERROR
        msg = "Can't restore ownership of \"";
        msg += (**p).name;
        msg += "\" : ";
        msg += strerror(errno);
        m->process_error(msg);
        }
    if (chmod((**p).name.c_str(), (**p).mode))
        {
        // ERROR
        msg = "Can't restore permissions of \"";
        msg += (**p).name;
        msg += "\" : ";
        msg += strerror(errno);
        m->process_error(msg);
        }
    utb.actime = (**p).atime;
    utb.modtime = (**p).mtime;
    if (utime( (char *)((**p).name.c_str()), &utb))
        {
        // ERROR
        msg = "Failed to restore datestamp of \""  + (**p).name;
        msg += "\" : ";
        msg += strerror(errno);
        m->process_error(msg);
        }
//     cout << "DONE " << n << endl;
    }
}

// Start file "errors" Restore Point / Restore Selection
bool Restore::start_errors_file(string *restore_point, char *type, SlaveMessages *m)
{
string errors_pname = *restore_point + "/errors";
errors_ofp.open(errors_pname.c_str(), ios::out | ios::binary | ios::app );
if ( ! errors_ofp.is_open() )
    {
      string msg = "Can't open \"" +  errors_pname;
      msg += "\" for append";
      m->process_error(msg);

      // COMPLAIN - ERROR - can't open error file
      // cerr << (char *)(STREQ(type, "Restore Point") ? "Restore" : type) << ":-" << msg << "\n";
      return FALSE;
    }

chmod(errors_pname.c_str(), 0644);
errors_ofp << "\n\n" << version_txt << "\nRESTORE ERRORS " << ctime(&m->start_time);
errors_ofp <<  (char *)(STREQ(type, "Restore Point") ? "Restore" : type) << " from:- " << *restore_point;
errors_ofp << "\nUser: ";
errors_ofp << (getpwuid( getuid() ))->pw_name;
errors_ofp << "\n";
m->put_error_fp(&errors_ofp);
return TRUE;
}

void Restore::get_compression(string *restore_point)
{
string name = *restore_point + "/gz";

if (isfile((char *)name.c_str()))
    {
    compression = "gz";
    }
else
    {
    name = *restore_point + "/bz2";

    if (isfile((char *)name.c_str()))
        {
        compression = "bz2";
        }
    else
        {
        name = *restore_point+ "/gz_bz2";

        if (isfile((char *)name.c_str()))
            {
            compression = "gz_bz2";
            }
        else compression = ""; // default
        }
    }
}
  
// ====================
// Restore from a restore point
// ====================
Restore::Restore(string *from, string *config, SlaveMessages *sm)
{
string pathname;
string d_name;
string s_name;
string errors_pname;
string msg;

// cout << "Restore::Restore(string *from, string *config, SlaveMessages *sm) - from a restore point\n";
  
data_pointer.clear();
m = sm;
m->num_names = 0;

if (config && config->length())
    {
    conf = new Conf(*config);
    }
else conf = new Conf();

umask(7777); // Make sure we can do anything...

// cout << "Restore(from) - from= " << *from << endl;


// If we have "from" then "from" is used to get the pathname
// otherwise it is obtained from conf
if (from == NULL || from->length() == 0)
    {
    // Get default restore source
    // from conf which itself might or might not be the default
    if (!find_backup(&pathname, conf, m))
        {
        // ERROR
        msg = "No source found for restore";

        // cout << "Restore(from): - " << msg << "\n";

        m->process_fatal_error(msg);
        return;
        }
    }
else  if (!get_parent_pathname(from, &pathname))
    {
    // ERROR
    msg = "Failed to get parent pathname";
//    cerr << "Restore: - " << msg << "\n";
    m->process_fatal_error(msg);

    return;
    }

msg = "Restore Point: " + pathname;
m->send_message(SL_POINT, (char *)msg.c_str());

if (!start_errors_file(&pathname, "Restore Point", m)) return;
get_compression(&pathname);

//cout << "Restore: - compression is " << compression << endl;

  // It appears we need the file data so that orphans can be fixed...
Backup::get_file_data(pathname, data, conf->restore_orphans, m);

s_name = pathname + "/root";
d_name =  "/";

// cout << "Restore: d_name= " << d_name << endl;
if (isdir((char *)s_name.c_str()))
    {
    process_directory(&d_name, &s_name);
    }
else  // Assume it is a file of some kind - the terminal filename is required
    {
    string filename;

    for (int n=s_name.length()-1; n>=0; --n)
        {
        if (s_name[n] == '/')
            {
            filename=s_name.substr(n);
            break;
            }
        }

    msg = "Failed to perform Restore: Directory \"";
    msg += s_name;
    msg += "\" does not exist";
    m->process_fatal_error(msg);
    return;
    }

fix_uid_etc();

//cout << "Restore: LEAVING" << endl;
}

// Add file data to the list of names
// Place the result into "data"
// Clear the selection list which is now in "data"
void Restore::add_file_data(string &restore_point)
{
string restore_name;
string root_name;
File_Data fd;
int p;
list<File_Data>::iterator i;
list<File_Data>::iterator j;

p = restore_point.length();

Backup::get_file_data(restore_point, data, TRUE, m);  // The whole list
root_name = restore_point + "/root";

i = m->restore_selection.begin();
if ( (*i).backup_name== root_name )
    {
    // We don't want this entry...
    m->restore_selection.pop_front();
    }
    
// Add the data to the list...
for (i=m->restore_selection.begin(); i!=m->restore_selection.end(); ++i)
    {
    if (isfile( (char *)(*i).backup_name.c_str() ) )
        {
        remove_compression_suffix(restore_name, (*i).backup_name, NULL, NULL);
         restore_name = restore_name.substr(root_name.length());
        }
    else restore_name = (*i).backup_name.substr(root_name.length());
    
        
    for (j=data.begin(); j!=data.end(); ++j)
        {            
        if ( (*j).name == restore_name)
            {                
            // Copy the data
            (*i).name = restore_name;
            (*i).uid = (*j).uid;
            (*i).gid = (*j).gid;
            (*i).mode = (*j).mode;
            (*i).atime = (*j).atime;
            (*i).mtime = (*j).mtime;
            break;
            }
        }
    }
    
data.clear();
for (i=m->restore_selection.begin(); i!=m->restore_selection.end(); ++i)
    {
    data.push_back(*i);
    }
m->restore_selection.clear();
}

// NOTE - we need configuration here, so pass it as a char * instead of string *
// ==================
// Restore a selection ONLY
// ==================
Restore::Restore(string *from, char *config, SlaveMessages *sm)
{
string s_name;
string d_name ;

string filename;
string directory;

string errors_pname ;
string msg;
list<File_Data>::iterator i;
string restore_point;

data_pointer.clear();
m = sm;

if (config && *config)
    {
    string c = config;
    conf = new Conf(c);
    }
  else conf = new Conf();

umask(7777); // Make sure we can do anything...

if ( (from == NULL) || (from->length() == 0) )
    {
    if (!find_backup(&restore_point, conf, m))
        {
        // ERROR
        msg = "No source found for restore";
        m->process_fatal_error(msg);
        return;
        }  
    }
else restore_point = *from;

msg = "Restore Point: " + restore_point;
m->send_message(SL_POINT, (char *)msg.c_str());

s_name = restore_point + "/root";
errors_pname = restore_point + "/errors";

if (!start_errors_file(&restore_point, "Restore Selection", m)) return;

get_compression(&restore_point);

// Place the data in "data"
add_file_data(restore_point);



// Traverse the list one item after another
// If a directory doesn't exist, create it
// Make the directories writeable
// Place all its contents (file, link, node)
// Traverse the list again to fix permissions

for (i=data.begin(); i!=data.end(); ++i)
    {
    s_name = (*i).backup_name;
    d_name = (*i).name;

    m->get_messages();
    if (m->quit_flag) break;

    if (isdir( (char*) s_name.c_str() ) )
        {
        if (!(isdir( (char*)d_name.c_str() ) ))
            {
            m->send_message(SL_PATH, (char *)d_name.c_str());

            // Create the directory
            if (!create_directory(&(d_name))) continue; // Try the rest of it anyway
          }
        Backup::change_permissions_to_0777(&(d_name), m);
        // Success so save the iterator!
        data_pointer.push_back(i);
        m->num_names++;
        }
    else  // Assume it is a file of some kind - the terminal filename is required
        {
        for (int n=d_name.length()-1; n>=0; --n)
            {
            if (d_name[n] == '/')
                {
                filename=d_name.substr(n);
                directory = d_name.substr(0, n);
                
//                cout << "Restore::Restore  - directory= " << directory << endl;
 //               cout << "                                - filename= " << filename << endl;
                
                break;
                }
            }
            
        m->send_message(SL_PATH, (char *)directory.c_str());
        m->send_message(SL_FILE, (char *)filename.c_str());
        
        if (isfile( (char*) s_name.c_str() ) )
            {
            if (!process_file(&d_name, &s_name))
                {
                // ERROR
                msg = "FAILED to restore file \"";
                msg += d_name;
                msg += "\"";
                m->process_error(msg);
                }    
            }
        else if (islink( (char*) s_name.c_str() ) )
            {
            if (!process_link(&d_name, &s_name))
                {
                // ERROR
                msg = "FAILED to restore link \"";
                msg += d_name;
                msg += "\"";
                m->process_error(msg);
                }
            }
        // Anything else is either a node or an error
        else if (!process_node(&d_name, &s_name))
            {
            // ERROR
            msg = "FAILED to restore special file \"";
            msg += d_name;
            msg += "\"";
            m->process_error(msg);
            }
        }
    }
    
// Fix ownership, permissions time etc.
fix_uid_etc();
}




